import logging
import warnings
from copy import deepcopy
from functools import partial
from typing import TYPE_CHECKING, Callable, Optional, Union

from mstrio import config
from mstrio.api import attributes, hierarchies, objects, tables
from mstrio.connection import Connection
from mstrio.modeling.expression import Expression, ExpressionFormat, FactExpression
from mstrio.modeling.schema.attribute import (
    AttributeForm,
    Relationship,
    RelationshipType,
)
from mstrio.modeling.schema.helpers import (
    AttributeDisplays,
    AttributeSorts,
    DataType,
    FormReference,
    ObjectSubType,
    SchemaObjectReference,
)
from mstrio.object_management import search_operations
from mstrio.object_management.folder import Folder
from mstrio.object_management.search_enums import SearchPattern
from mstrio.types import ObjectSubTypes, ObjectTypes
from mstrio.utils.entity import CopyMixin, DeleteMixin, Entity, MoveMixin
from mstrio.utils.enum_helper import get_enum_val
from mstrio.utils.helper import (
    delete_none_values,
    filter_params_for_func,
    get_valid_project_id,
)
from mstrio.utils.version_helper import class_version_handler, method_version_handler

if TYPE_CHECKING:
    from mstrio.modeling.schema.attribute import Attribute

logger = logging.getLogger(__name__)


@method_version_handler('11.3.0100')
def list_attributes(
    connection: Connection,
    name: Optional[str] = None,
    attribute_subtype: Optional[ObjectSubTypes] = None,
    to_dictionary: bool = False,
    limit: Optional[int] = None,
    search_pattern: Union[SearchPattern, int] = SearchPattern.CONTAINS,
    project_id: Optional[str] = None,
    project_name: Optional[str] = None,
    show_expression_as: Union[ExpressionFormat, str] = ExpressionFormat.TREE,
    **filters,
) -> Union[list["Attribute"], list[dict]]:
    """Get list of Attribute objects or dicts with them.
    Optionally filter attributes by specifying 'name', 'attribute_subtype'.

    Optionally use `to_dictionary` to choose output format.

    Wildcards available for 'name':
        ? - any character
        * - 0 or more of any characters
        e.g. name_begins = ?onny will return Sonny and Tonny

    Specify either `project_id` or `project_name`.
    When `project_id` is provided (not `None`), `project_name` is omitted.

    Note:
        When `project_id` is `None` and `project_name` is `None`,
        then its value is overwritten by `project_id` from `connection` object.

    Args:
        connection: MicroStrategy connection object returned by
            `connection.Connection()`
        name (string, optional): value the search pattern is set to, which
            will be applied to the names of attributes being searched
        attribute_subtype (ObjectSubTypes): one of attribute subtypes:
            attribute, attribute_abstract, attribute_recursive, attribute_role,
            attribute_transformation
        to_dictionary (bool, optional): If True returns dict, by default (False)
            returns Attribute objects
        limit (integer, optional): limit the number of elements returned. If
            None all object are returned.
        project_id (str, optional): Project ID
        project_name (str, optional): Project name
        search_pattern (SearchPattern enum or int, optional): pattern to search
            for, such as Begin With or Exactly. Possible values are available in
            ENUM mstrio.object_management.SearchPattern.
            Default value is CONTAINS (4).
        show_expression_as (ExpressionFormat, str): specify how expressions
            should be presented
            Available values:
                - `ExpressionFormat.TREE` or `tree` (default)
                - `ExpressionFormat.TOKENS or `tokens`
        **filters: Available filter parameters:
            id str: Attribute's ID
            name str: Attribute's name
            description str: Attribute's description
            date_created str: format: 2001-01-02T20:48:05.000+0000
            date_modified str: format: 2001-01-02T20:48:05.000+0000
            version str: Attribute's version
            owner dict: e.g. {'id': <user's id>, 'name': <user's name>},
                with one or both of the keys: id, name
            acg str | int: access control group
            subtype str: object's subtype
            ext_type str: object's extended type

    Returns:
        list with Attribute objects or list of dictionaries
    """
    project_id = get_valid_project_id(
        connection=connection,
        project_id=project_id,
        project_name=project_name,
        with_fallback=False if project_name else True,
    )

    if attribute_subtype is None:
        attribute_subtype = ObjectTypes.ATTRIBUTE
    objects_ = search_operations.full_search(
        connection,
        object_types=attribute_subtype,
        project=project_id,
        name=name,
        pattern=search_pattern,
        limit=limit,
        **filters,
    )
    if to_dictionary:
        return objects_
    else:
        show_expression_as = (
            show_expression_as
            if isinstance(show_expression_as, ExpressionFormat)
            else ExpressionFormat(show_expression_as)
        )
        return [
            Attribute.from_dict(
                {**obj_, 'show_expression_as': show_expression_as}, connection
            )
            for obj_ in objects_
        ]


@class_version_handler('11.3.0100')
class Attribute(Entity, CopyMixin, MoveMixin, DeleteMixin):  # noqa
    """Python representation of MicroStrategy Attribute object.

    Attributes:
        id: attribute's ID
        name: attribute's name
        sub_type: string literal used to identify the type of a metadata object
        description: attribute's description
        type: object type, ObjectTypes enum
        subtype: object subtype, ObjectSubTypes enum
        ext_type: object extended type, ExtendedType enum
        ancestors: list of ancestor folders
        date_created: creation time, DateTime object
        date_modified: last modification time, DateTime object
        destination_folder_id: a globally unique identifier used to distinguish
            between metadata objects within the same project
        forms: the list of attribute forms
        attribute_lookup_table: Information about an object referenced within
            the specification of another object. An object reference typically
            contains only enough fields to uniquely identify the referenced
            objects.
        key_form: a key form of an attribute
        displays: The collections of attribute displays and browse displays
        sorts: the collections of attribute sorts and browse sorts
            of the attribute.
        relationships: the list of relationships that one attribute has.
        is_embedded: If true indicates that the target object of this
            reference is embedded within this object. Alternatively if
            this object is itself embedded, then it means that the target
            object is embedded in the same container as this object.
        owner: User object that is the owner
        acg: access rights (See EnumDSSXMLAccessRightFlags for possible values)
        acl: object access control list
        hidden: Specifies whether the object is hidden.
    """

    _OBJECT_TYPE = ObjectTypes.ATTRIBUTE
    _API_GETTERS = {
        (
            'id',
            'sub_type',
            'name',
            'is_embedded',
            'description',
            'destination_folder_id',
            'forms',
            'attribute_lookup_table',
            'key_form',
            'displays',
            'sorts',
            'relationships',
        ): attributes.get_attribute,
        (
            'abbreviation',
            'type',
            'subtype',
            'ext_type',
            'date_created',
            'date_modified',
            'version',
            'owner',
            'icon_path',
            'view_media',
            'ancestors',
            'certified_info',
            'acg',
            'acl',
            'target_info',
            'hidden',
        ): objects.get_object_info,
    }
    _API_PATCH = {
        (
            'id',
            'sub_type',
            'name',
            'is_embedded',
            'description',
            'destination_folder_id',
            'forms',
            'attribute_lookup_table',
            'key_form',
            'displays',
            'sorts',
        ): (attributes.update_attribute, 'partial_put'),
        'relationships': (hierarchies.update_attribute_relationships, 'partial_put'),
        ('folder_id', 'hidden'): (objects.update_object, 'partial_put'),
    }
    _FROM_DICT_MAP = {
        **Entity._FROM_DICT_MAP,
        "forms": (
            lambda source, connection: [
                AttributeForm.from_dict(content, connection) for content in source
            ]
        ),
        "relationships": (
            lambda source, connection: [
                Relationship.from_dict(content, connection) for content in source
            ]
        ),
        "attribute_lookup_table": SchemaObjectReference.from_dict,
        "key_form": FormReference.from_dict,
        "displays": AttributeDisplays.from_dict,
        "sorts": AttributeSorts.from_dict,
    }

    @staticmethod
    def validate_key_form(
        key_form: FormReference,
        forms: list[AttributeForm],
        error_msg: Optional[str] = None,
    ) -> FormReference:
        """Validate whether the key form exists in the list of attribute forms
            provided

        Args:
            key_form: a key form of an attribute to perform validation on
            forms: the list of the attribute forms
            error_msg (optional): optional message to display instead of the
                standard one

        Returns:
            A validated key form."""

        # if forms is None or len(forms) < 1:
        if not forms:
            raise AttributeError("`forms` can not be empty.")
        elif len(forms) == 1:
            return FormReference(name=forms[0].name)
        elif key_form is None or not any(
            form.is_referenced_by(key_form) for form in forms
        ):
            raise AttributeError(
                error_msg or "Please select a `key_form` from the `forms` provided."
            )
        return key_form

    @staticmethod
    def check_if_referenced_forms_exist(
        error_msg: str, forms: list[AttributeForm], refs: list[FormReference]
    ):
        """Check if all references point to a form in forms."""
        for ref in refs:
            flag = False
            for form in forms:
                if form.is_referenced_by(ref):
                    flag = True
            if flag is False:
                raise AttributeError(error_msg)

    @staticmethod
    def validate_displays(
        displays: AttributeDisplays, forms: list[AttributeForm]
    ) -> AttributeDisplays:
        """Validate whether the Attribute Displays are populated correctly and
            only use references to forms present in the Attribute Forms of this
            particular attribute.

        Args:
            displays: AttributeDisplays of the Attribute
            forms: list of AttributeForm objects of the Attribute

        Returns:
            Validated, non-empty and properly referencing Attribute Displays"""
        # Displays CAN'T be empty. Create and populate with form refs
        if displays is None:
            form_refs = [FormReference(id=form.id) for form in forms]
            return AttributeDisplays(
                report_displays=form_refs, browse_displays=form_refs
            )
        if displays.report_displays == []:
            displays.report_displays = [FormReference(id=form.id) for form in forms]
        if displays.browse_displays == []:
            displays.browse_displays = [FormReference(id=form.id) for form in forms]

        # Validate if displays use form refs not present in forms
        error_msg = "FormReference present in `displays` is not present in `forms`."
        Attribute.check_if_referenced_forms_exist(
            error_msg, forms, displays.report_displays
        )
        Attribute.check_if_referenced_forms_exist(
            error_msg, forms, displays.browse_displays
        )
        return displays

    @staticmethod
    def validate_sorts(
        sorts: AttributeSorts, forms: list[AttributeForm]
    ) -> Optional[AttributeSorts]:
        """Validate whether the sorts use form references that aren't present
            in the provided forms

        Args:
            sorts: the collections of attribute sorts and browse sorts
                of the attribute
            forms: list of AttributeForm objects of the Attribute

        Returns:
            Validated sorts or nothing if the provided sorts were empty."""
        if sorts is None or (sorts.browse_sorts is None and sorts.report_sorts is None):
            return None

        # Validate if sorts use form refs not present in forms
        error_msg = "FormReference present in `sort` is not present in `forms`."
        if sorts.report_sorts is not None:
            Attribute.check_if_referenced_forms_exist(
                error_msg, forms, [attr_sort.form for attr_sort in sorts.report_sorts]
            )
        if sorts.browse_sorts is not None:
            Attribute.check_if_referenced_forms_exist(
                error_msg, forms, [attr_sort.form for attr_sort in sorts.browse_sorts]
            )
        return sorts

    @classmethod
    def create(
        cls,
        connection: 'Connection',
        name: str,
        sub_type: Union[ObjectSubType, str],
        destination_folder: Union[Folder, str],
        forms: list[AttributeForm],
        key_form: FormReference,
        displays: AttributeDisplays,
        description: Optional[str] = None,
        is_embedded: bool = False,
        attribute_lookup_table: Optional[SchemaObjectReference] = None,
        sorts: Optional[AttributeSorts] = None,
        show_expression_as: Union[ExpressionFormat, str] = ExpressionFormat.TREE,
    ) -> "Attribute":
        """Alter attribute properties.

        Args:
            connection: MicroStrategy connection object returned
                by `connection.Connection()`
            name: attribute's name
            sub_type: attribute's sub_type
            destination_folder_id: A globally unique identifier used to
                distinguish between metadata objects within the same project.
                It is possible for two metadata objects in different projects
                to have the same Object Id.
            forms: attribute's forms list
            key_form: a key form of an attribute
            displays: The collections of attribute displays and browse displays
                of the attribute.
            description: attribute's description
            is_embedded: If true indicates that the target object of this
                reference is embedded within this object. Alternatively if
                this object is itself embedded, then it means that the target
                object is embedded in the same container as this object.
            attribute_lookup_table: Information about an object referenced
                within the  specification of another object. An object reference
                typically contains only enough fields to uniquely identify
                the referenced objects.
            sorts: The collections of attribute sorts and browse sorts
                of the attribute.
            show_expression_as (ExpressionFormat, str): specify how expressions
                should be presented
                Available values:
                - `ExpressionFormat.TREE` or `tree` (default)
                - `ExpressionFormat.TOKENS or `tokens`

        Returns:
            Attribute class object.
        """
        # Validate dependencies on forms
        key_form = cls.validate_key_form(key_form=key_form, forms=forms)
        displays = cls.validate_displays(displays=displays, forms=forms)
        sorts = cls.validate_sorts(sorts, forms)

        body = {
            'information': {
                'name': name,
                'subType': get_enum_val(sub_type, ObjectSubType),
                'isEmbedded': is_embedded,
                'description': description,
                'destinationFolderId': destination_folder.id
                if isinstance(destination_folder, Folder)
                else destination_folder,
            },
            'forms': [form.to_dict() for form in forms] if forms else None,
            'attributeLookupTable': attribute_lookup_table.to_dict()
            if attribute_lookup_table
            else None,
            'keyForm': key_form.to_dict() if key_form else None,
            'displays': displays.to_dict() if displays else None,
            'sorts': sorts.to_dict() if sorts else None,
        }
        body = delete_none_values(body, recursion=True)
        response = attributes.create_attribute(
            connection,
            body=body,
            show_expression_as=get_enum_val(show_expression_as, ExpressionFormat),
        ).json()

        if config.verbose:
            logger.info(
                f"Successfully created attribute named: '{name}' with ID: '"
                f"{response['id']}'"
            )

        return cls.from_dict(
            source={**response, 'show_expression_as': show_expression_as},
            connection=connection,
        )

    def __init__(
        self,
        connection: Connection,
        id: Optional[str] = None,
        name: Optional[str] = None,
        show_expression_as: Union[ExpressionFormat, str] = ExpressionFormat.TREE,
    ) -> None:
        """Initializes a new instance of Attribute class

        Args:
            connection (Connection): MicroStrategy connection object returned
                by `connection.Connection()`
            id (str, optional): Attribute's ID. Defaults to None.
            name (str, optional): Attribute's name. Defaults to None.
            show_expression_as (ExpressionFormat or str, optional):
                specify how expressions should be presented.
                Defaults to ExpressionFormat.TREE.

                Available values:
                - `ExpressionFormat.TREE` or `tree` (default)
                - `ExpressionFormat.TOKENS` or `tokens`
        Note:
            Parameter `name` is not used when fetching. If only `name` parameter
            is provided, `id` will be found automatically if such object exists.

        Raises:
            ValueError: if both `id` and `name` are not provided or
            if Attribute with the given `name` doesn't exist.
        """
        if id is None:
            attribute = super()._find_object_with_name(
                connection=connection,
                name=name,
                listing_function=partial(
                    list_attributes, search_pattern=SearchPattern.EXACTLY
                ),
            )
            id = attribute['id']
        super().__init__(
            connection=connection,
            object_id=id,
            name=name,
            show_expression_as=show_expression_as,
        )

    def _init_variables(self, **kwargs) -> None:
        super()._init_variables(**kwargs)
        self._sub_type = kwargs.get('sub_type')
        self._is_embedded = kwargs.get('is_embedded')
        self._destination_folder_id = kwargs.get('destination_folder_id')
        self._forms = (
            [
                AttributeForm.from_dict(expr, self._connection)
                for expr in kwargs.get('forms')
            ]
            if kwargs.get('forms')
            else None
        )
        self._attribute_lookup_table = (
            SchemaObjectReference.from_dict(kwargs.get('attribute_lookup_table'))
            if kwargs.get('attribute_lookup_table')
            else None
        )
        self._key_form = (
            FormReference.from_dict(kwargs.get('key_form'))
            if kwargs.get('key_form')
            else None
        )
        self._displays = (
            AttributeDisplays.from_dict(kwargs.get('displays'))
            if kwargs.get('displays')
            else None
        )
        self._sorts = (
            AttributeSorts.from_dict(kwargs.get('sorts'))
            if kwargs.get('sorts')
            else None
        )
        self._relationships = (
            [kwargs.get('relationships')] if kwargs.get('relationships') else None
        )
        show_expression_as = kwargs.get('show_expression_as', 'tree')
        self._show_expression_as = (
            show_expression_as
            if isinstance(show_expression_as, ExpressionFormat)
            else ExpressionFormat(show_expression_as)
        )

    def alter(
        self,
        sub_type: Optional[str] = None,
        name: Optional[str] = None,
        is_embedded: Optional[bool] = None,
        description: Optional[str] = None,
        destination_folder_id: Optional[str] = None,
        forms: Optional[list[AttributeForm]] = None,
        attribute_lookup_table: Optional[SchemaObjectReference] = None,
        key_form: Optional[FormReference] = None,
        displays: Optional[AttributeDisplays] = None,
        sorts: Optional[AttributeSorts] = None,
        relationships: Optional[Relationship] = None,
        hidden: Optional[bool] = None,
    ):
        """Alter attribute properties.

        Args:
            name: attribute's name
            description: attribute's description
            is_embedded: If true indicates that the target object of this
                reference is embedded within this object. Alternatively if
                this object is itself embedded, then it means that the target
                object is embedded in the same container as this object.
            destination_folder_id: A globally unique identifier used to
                distinguish between metadata objects within the same project.
                It is possible for two metadata objects in different projects
                to have the same Object Id.
            forms: attribute's forms list
            attribute_lookup_table: Information about an object referenced
                within the  specification of another object. An object reference
                typically contains only enough fields to uniquely identify
                the referenced objects.
            key_form: a key form of an attribute
            displays: The collections of attribute displays and browse displays
                of the attribute.
            sorts: The collections of attribute sorts and browse sorts
                of the attribute.
            relationships: the list of relationships that one attribute has.
            hidden: Specifies whether the attribute is hidden.
        """
        # REST doesn't return hidden attribute if its value is False.
        # If attribute is not present in response, the mstrio engine
        # interprets this as 'no value' and sets its value to None.
        # But if attribute's value is not None, and REST returns nothing
        # the engine does nothing, and doesn't change its value.
        # So if 'hidden' has value True, and is changed to False,
        # the REST will return nothing, and locally its value will stay True.
        # This line is to update local value of 'hidden'.
        self._hidden = hidden

        key_form = self.validate_key_form(
            key_form or self.key_form, forms or self.forms
        )
        displays = self.validate_displays(
            displays or self.displays, forms or self.forms
        )
        sorts = self.validate_sorts(sorts or self.sorts, forms or self.forms)

        properties = filter_params_for_func(self.alter, locals(), exclude=['self'])
        self._alter_properties(**properties)

    # Attribute relationships management
    def add_child(
        self,
        child: Optional[SchemaObjectReference] = None,
        joint_child: Optional[list[SchemaObjectReference]] = None,
        relationship_type: RelationshipType = RelationshipType.ONE_TO_MANY,
        table: Optional[SchemaObjectReference] = None,
    ) -> None:
        """Add a child to the attribute.

        Args:
            child: SchemaObjectReference of an attribute
            joint_child: list of SchemaObjectReferences of an attributes
            relationship_type: RelationshipType enum object, by default
                RelationshipType.ONE_TO_MANY
            table: SchemaObjectReference of a lookup table, if not passed
                attribute lookup table is used
        """
        if (child and joint_child) or (not child and not joint_child):
            raise ValueError(
                "Please specify either 'child' or 'joint_child' parameter."
            )
        elif child:
            for rel in self.relationships:
                if hasattr(rel, 'child') and rel.child == child:
                    warnings.warn(
                        f"{child.name} already is a child of the attribute '{self.id}'"
                        " and will be omitted."
                    )
                    return None
        elif joint_child:
            for rel in self.relationships:
                if hasattr(rel, 'joint_child') and rel.joint_child == joint_child:
                    children = (
                        "[" + ", ".join(child.name for child in joint_child) + "]"
                    )
                    warnings.warn(
                        f"{children} already is a joint_child of the attribute '"
                        f"{self.id}' and will be omitted."
                    )
                    return None

        parent = SchemaObjectReference.create_from(self)
        table = table or self.attribute_lookup_table

        return self._update_relationships(
            Relationship(relationship_type, table, parent, child, joint_child)
        )

    def add_parent(
        self,
        parent: SchemaObjectReference,
        relationship_type: RelationshipType = RelationshipType.ONE_TO_MANY,
        table: Optional[SchemaObjectReference] = None,
    ) -> None:
        """Add a parent to the attribute.

        Args:
            parent: SchemaObjectReference of an attribute
            relationship_type: RelationshipType enum object, by default
                RelationshipType.ONE_TO_MANY
            table: SchemaObjectReference of a lookup table, if not passed
                attribute lookup table is used
        """
        for rel in self.relationships:
            if rel.parent.object_id == parent.object_id:
                warnings.warn(
                    f"{parent.name} already is a parent of the attribute '{self.id}'"
                    " and will be omitted."
                )
                return None
        child = SchemaObjectReference.create_from(self)
        table = table or self.attribute_lookup_table

        return self._update_relationships(
            Relationship(relationship_type, table, parent, child)
        )

    def remove_child(
        self,
        child: Optional[SchemaObjectReference] = None,
        joint_child: Optional[list[SchemaObjectReference]] = None,
    ) -> None:
        """Removes a child of the attribute.

        Args:
            child: SchemaObjectReference of an attribute to be removed
                from child relationship
            joint_child: list of SchemaObjectReferences of an attributes
                to be removed from joint child relationship
        """
        if (child and joint_child) or (not child and not joint_child):
            raise ValueError(
                "Please specify either 'child' or 'joint_child' parameter."
            )
        elif joint_child:
            for rel in self.relationships:
                if hasattr(rel, 'joint_child') and rel.joint_child == joint_child:
                    return self._update_relationships(rel, add=False)
            child_name = "[" + ", ".join(child.name for child in joint_child) + "]"
        elif child:
            for rel in self.relationships:
                if hasattr(rel, 'child') and rel.child == child:
                    return self._update_relationships(rel, add=False)
            child_name = child.name

        warnings.warn(
            f"{child_name} is not a child/joint_child of the attribute '{self.id}'"
            " and will be omitted."
        )

    def remove_parent(self, parent: SchemaObjectReference) -> None:
        """Removes a parent of the attribute.

        Args:
            parent: SchemaObjectReference of an attribute to be removed
                from parent relationship
        """
        for rel in self.relationships:
            if rel.parent.object_id == parent.object_id and parent.object_id != self.id:
                return self._update_relationships(rel, add=False)

        warnings.warn(
            f"{parent.name} is not a parent of the attribute '{self.id}' and will be "
            f"omitted."
        )

    def _update_relationships(self, relationship: Relationship, add=True) -> None:
        """Inner method for sending updated relationships list.

        Args:
            relationship: relationship to add or remove
            add: if True the relationship will be added to the attribute
                relationships, else it will be removed from them
        """
        relationships = self.relationships.copy()
        if add:
            relationships.append(relationship)
        else:
            relationships = [rel for rel in relationships if rel != relationship]
        self._alter_properties(relationships=relationships)

    def list_relationship_candidates(
        self, already_used: bool = True, to_dictionary: bool = True
    ) -> Union[dict, list[SchemaObjectReference]]:
        """Lists potential relationship candidates for the Attribute.

        Args:
            already_used: whether to show Attributes References, which
                are already parents or children
            to_dictionary: whether to return results as a dict or list
        Returns:
            Dictionary with table names as keys and list
            of SchemaObjectReferences of the attributes as values
            if to_dictionary set to True, list of SchemaObjectReference
            of attributes as list otherwise.
        """
        key_form_expressions = [
            expr
            for form in self.forms
            if form.id == self.key_form.id
            for expr in form.expressions
        ]
        potential_tables = [tab for exp in key_form_expressions for tab in exp.tables]

        result = {}
        for tab in potential_tables:
            table = tables.get_table(
                self.connection, tab.object_id, project_id=self.connection.project_id
            )
            attribute_references = table.json()['attributes']

            candidates = []
            for attr in attribute_references:
                attr_copy = attr.copy()
                if attr_copy.get('id'):
                    attr_copy.update({'objectId': attr_copy.pop('id')})
                if attr_copy.get('information'):
                    attr_copy.update(attr_copy.pop('information'))
                if attr_copy.get('objectId') != self.id:
                    candidates.append(SchemaObjectReference.from_dict(attr_copy))
            if candidates:
                result[tab.name] = candidates

        if already_used is False:
            children = [
                rel.child
                for rel in self.relationships
                if hasattr(rel, 'child') and rel.child.object_id != self.id
            ]
            parents = [
                rel.parent
                for rel in self.relationships
                if rel.parent.object_id != self.id
            ]
            used = [*children, *parents]
            result = {
                tab: [candidate for candidate in candidates if candidate not in used]
                for tab, candidates in result.items()
            }

        if to_dictionary is False:
            result = list(
                {ref for references in list(result.values()) for ref in references}
            )

        return result

    def list_tables(
        self, expression: Optional[Union[FactExpression, str]] = None
    ) -> list[SchemaObjectReference]:
        """List all tables in the given expression. If expression is not
        specified, list all tables for attribute.

        Args:
            expression: the expression for which to list tables

        Returns:
            List of tables in the given expression or all tables for attribute.
        """
        expressions = [
            expr for form in self.forms if form.expressions for expr in form.expressions
        ]
        if expression:
            expression_id = (
                expression.id if isinstance(expression, FactExpression) else expression
            )
            expressions = [expr for expr in expressions if expr.id == expression_id]

        table_list = {tab for expr in expressions for tab in expr.tables}

        return list(table_list)

    # Attribute forms management
    def __alter_form_with_id(self, form_id: str, func: Callable, params: dict):
        forms = []
        is_form_found = False

        for form in self.forms:
            if form.id == form_id:
                is_form_found = True
                func(form, **params)
            forms.append(form)

        if not is_form_found:
            raise ValueError(f"Attribute Form with ID: {form_id} has not been found.")

        self._alter_properties(forms=forms)

    def get_form(self, id: str = None, name: str = None) -> Optional[AttributeForm]:
        """Retrieve a certain attribute form of a local instance of
        Attribute object.

        Args:
            id: ID of the attribute form. It has priority over `name` parameter
            name: name of the attribute form
        """
        if id:
            for form in self.forms:
                if form.id == id:
                    return form
        elif name:
            for form in self.forms:
                if form.name == name:
                    return form
        else:
            raise ValueError("Provide id or name.")

    def get_fact_expression(
        self, expression_id: str, form_id: str = None, form_name: str = None
    ) -> Optional[FactExpression]:
        """Retrieve a certain fact expression of a local instance of
        Attribute object.

        Args:
            form_id: ID of the attribute form. It have priority over `name`
                parameter
            form_name: name of the attribute form
            expression_id: ID of the fact expression
        """
        form = self.get_form(form_id, form_name)
        return form.get_fact_expression(expression_id)

    def add_form(
        self,
        form: Optional[AttributeForm] = None,
        name: Optional[str] = None,
        expressions: Optional[list[FactExpression]] = None,
        lookup_table: Optional[SchemaObjectReference] = None,
        description: Optional[str] = None,
        category: Optional[str] = None,
        display_format: Optional[AttributeForm.DisplayFormat] = None,
        data_type: Optional[DataType] = None,
        alias: Optional[str] = None,
        child_forms: Optional[list[FormReference]] = None,
        geographical_role: Optional[AttributeForm.GeographicalRole] = None,
        time_role: Optional[AttributeForm.TimeRole] = None,
        is_form_group: bool = False,
        is_multilingual: bool = False,
    ):
        """Create new attribute form and add it to the `attribute.forms` list.
        The form can be added from completed `AttributeForm` objects provided
        in `form` parameter, or by filling other parameters.

        Args:
            form: complete `AttibuteForm` object that will be added to
                the Attribute
            name: The name of the attribute form set by the attribute. Unlike
                category, which is the systemic name associated with each
                reusable form, this name is specific to the attribute using this
                form
            expressions: Array with a member object for each separately defined
                expression currently in use by a fact. Often a fact expression
                takes the form of just a single column name, but more complex
                expressions are possible.
            lookup_table: lookup table of the AttributeForm. It has to be a
                lookup table used in one of the expressions assigned to
                AttributeForm
            description: description of the AttributeForm
            category: The category of the attribute form. Unlike name, this
                field is independent of the attribute using this form. This
                field can only be set when creating a new form. Once a form
                is created, its category becomes non-mutable. If not provided
                (or set as None) when an attribute is being created, a custom
                category will be automatically generated.
            display_format: display format of the AttributeForm
            data_type: Representation in the object model for a data-type that
                could be used for a SQL column.
            alias: alias of the AttributeForm
            child_forms: only used if 'is_form_group' is set to true
            geographical_role: identifies the type of geographical information
                this form represents
            time_role: time role of the AttributeForm
            is_form_group: A boolean field indicating whether this form is
                a form group (if true) or a simple form (if false).
            is_multilingual: A boolean field indicating whether this field is
                multilingual. Any key form of the attribute is not allowed
                to be set as multilingual.
        """
        if form and isinstance(form, AttributeForm):
            self._alter_properties(forms=self.forms + [form])
        elif expressions and lookup_table:
            properties = filter_params_for_func(
                AttributeForm.local_create, locals(), exclude=['self']
            )
            properties = delete_none_values(properties, recursion=True)
            new_form = AttributeForm.local_create(
                connection=self.connection, **properties
            )
            self._alter_properties(forms=self.forms + [new_form])
        else:
            raise AttributeError(
                'Please provide either `form` or `expressions` and `lookup_table`'
            )

    @staticmethod
    def _remove_form_from_displays(
        form_to_be_removed: AttributeForm,
        forms: list[AttributeForm],
        displays: AttributeDisplays,
    ) -> AttributeDisplays:
        """Remove all references to the form from local instance of displays,
        AttributeDisplays object.

        Args:
            form_to_be_removed: form that will be removed
            forms: Attribute.forms list without the form that will be
                removed
            displays: Attribute.displays from which the form will be removed

        Return:
            AttributeDisplays object without references to the removed form
        """
        displays = deepcopy(displays)
        for index in range(len(displays.report_displays)):
            if form_to_be_removed.is_referenced_by(displays.report_displays[index]):
                displays.report_displays.pop(index)
                break
        for index in range(len(displays.browse_displays)):
            if form_to_be_removed.is_referenced_by(displays.browse_displays[index]):
                displays.browse_displays.pop(index)
                break
        return Attribute.validate_displays(displays, forms)

    @staticmethod
    def _remove_form_from_sorts(
        form_to_be_removed: AttributeForm,
        forms: list[AttributeForm],
        sorts: AttributeSorts,
    ) -> AttributeSorts:
        """Remove all references to the form from local instance of displays,
        AttributeSorts object.

        Args:
            form_to_be_removed: form that will be removed
            forms: Attribute.forms list without the form that will be
                removed
            sorts: Attribute.sorts from which the form will be removed

        Return:
            AttributeSorts object without references to the removed form
        """
        sorts = deepcopy(sorts)
        if sorts.report_sorts:
            for index in range(len(sorts.report_sorts)):
                if form_to_be_removed.is_referenced_by(sorts.report_sorts[index].form):
                    sorts.report_sorts.pop(index)
                    break
        if sorts.browse_sorts:
            for index in range(len(sorts.browse_sorts)):
                if form_to_be_removed.is_referenced_by(sorts.browse_sorts[index].form):
                    sorts.browse_sorts.pop(index)
                    break
        return Attribute.validate_sorts(sorts, forms)

    def remove_form(self, form_id: str, new_key_form: Optional[FormReference] = None):
        """Remove attribute form with a given `form_id`. If this form was
        present in `displays` or `sorts`, it will be automatically removed
        from there.

        Args:
            form_id: ID of the form that should be removed
            new_key_form: new value of the attribute's `key_form` parameter.
                Mandatory if the form that is to be removed is the current
                key form, otherwise optional.
        """
        new_forms = []
        form_to_del = None
        for form in self.forms:
            if form.id == form_id:
                form_to_del = form
            else:
                new_forms.append(form)

        # Make sure the key form is changed if needed, and other basic logic
        new_key_form = new_key_form or self.key_form
        if form_to_del is None:
            raise ValueError(
                f'Attribute with ID {self.id} does not '
                f'contain attribute form with ID {form_id}.'
            )
        elif len(new_forms) == 0:
            raise ValueError('You can not delete the last attribute form')
        elif form_to_del.is_referenced_by(self.key_form):
            error_msg = (
                'You are trying to delete the current key form. Please chose '
                'new key form.'
            )
            # Assign new key form if possible
            new_key_form = self.validate_key_form(new_key_form, new_forms, error_msg)

        # Make sure the form that is also removed from displays and sorts
        displays = self._remove_form_from_displays(
            form_to_del, new_forms, self.displays
        )
        sorts = self._remove_form_from_sorts(form_to_del, new_forms, self.sorts)

        self._alter_properties(
            key_form=new_key_form, forms=new_forms, displays=displays, sorts=sorts
        )

    def alter_form(
        self,
        form_id: str,
        name: Optional[str] = None,
        description: Optional[str] = None,
        display_format: Optional[AttributeForm.DisplayFormat] = None,
        data_type: Optional[DataType] = None,
        expressions: Optional[list[FactExpression]] = None,
        alias: Optional[str] = None,
        lookup_table: Optional[SchemaObjectReference] = None,
        child_forms: Optional[list[FormReference]] = None,
        geographical_role: Optional[AttributeForm.GeographicalRole] = None,
        time_role: Optional[AttributeForm.TimeRole] = None,
        is_form_group: Optional[bool] = None,
        is_multilingual: Optional[bool] = None,
    ):
        """Alter attribute form with a given `form_id`.

        Args:
            form_id: ID of the attribute form that will be altered
            name: The name of the attribute form set by the attribute. Unlike
                category, which is the systemic name associated with each
                reusable form, this name is specific to the attribute using
                this form
            description: description of the AttributeForm
            display_format: display format of the AttributeForm
            data_type: Representation in the object model for a data-type that
                could be used for a SQL column.
            expressions: Array with a member object for each separately defined
                expression currently in use by a fact. Often a fact expression
                takes the form of just a single column name, but more complex
                expressions are possible.
            alias: alias of the AttributeForm
            lookup_table: lookup table of the AttributeForm. It has to be a
                lookup table used in one of the expressions assigned to
                AttributeForm
            child_forms: only used if 'is_form_group' is set to true
            geographical_role: identifies the type of geographical information
                this form represents
            time_role: time role of the AttributeForm
            is_form_group: A boolean field indicating whether this form is
                a form group (if true) or a simple form (if false).
            is_multilingual: A boolean field indicating whether this field is
                multilingual. Any key form of the attribute is not allowed
                to be set as multilingual."""
        form_properties = filter_params_for_func(
            AttributeForm.local_alter, locals(), exclude=['self']
        )
        self.__alter_form_with_id(form_id, AttributeForm.local_alter, form_properties)

    # Attribute form expressions management
    def alter_fact_expression(
        self,
        form_id: str,
        fact_expression_id: str,
        expression: Optional['Expression'] = None,
        tables: Optional[list[SchemaObjectReference]] = None,
    ):
        """Alter fact expression of the attribute form with given ID
        Args:
            form_id: ID of the form that uses certain expression
            fact_expression_id: ID of the fact expression that is to be altered,
            expression: new expressions of the fact expression
            tables: new tables of the fact expression
        """
        expression_properties = filter_params_for_func(
            AttributeForm._alter_expression, locals(), exclude=['self']
        )
        self.__alter_form_with_id(
            form_id,
            AttributeForm._alter_expression,
            {'fact_expression_id': fact_expression_id, **expression_properties},
        )

    def add_fact_expression(self, form_id: str, expression: FactExpression):
        """Add expression to the form.
        Args:
            form_id: ID of the form to which the expression is to be added,
            expression: the expression that is to be added,
        """
        self.__alter_form_with_id(
            form_id, AttributeForm._add_fact_expression, {'expression': expression}
        )

    def remove_fact_expression(
        self,
        form_id: str,
        fact_expression_id: str,
        new_lookup_table: Optional[SchemaObjectReference] = None,
    ):
        """Remove expression from the form. If the expressions left are
        not using lookup table assigned to the form, provide new lookup
        table for the form.

        Args:
            form_id: ID of the form from which the expression is to be removed,
            expression_id: ID of the expression that is to be removed,
            new_lookup_table: new lookup table of the form
        """
        self.__alter_form_with_id(
            form_id,
            AttributeForm._remove_fact_expression,
            {
                'fact_expression_id': fact_expression_id,
                'new_lookup_table': new_lookup_table,
            },
        )

    def to_dict(self, camel_case: bool = True) -> dict:
        result = super().to_dict(camel_case)
        result.pop('_showExpressionAs', None)

        return result

    @property
    def sub_type(self):
        return self._sub_type

    @property
    def is_embedded(self):
        return self._is_embedded

    @property
    def destination_folder_id(self):
        return self._destination_folder_id

    @property
    def forms(self):
        return self._forms

    @property
    def attribute_lookup_table(self):
        return self._attribute_lookup_table

    @property
    def key_form(self):
        return self._key_form

    @property
    def displays(self):
        return self._displays

    @property
    def sorts(self):
        return self._sorts

    @property
    def relationships(self):
        return self._relationships

    @property
    def hidden(self):
        return self._hidden or False



import functools
import logging

from requests_futures.sessions import FuturesSession


class FuturesSessionWithRenewal(FuturesSession):
    def __init__(self, *, connection, **kwargs):
        super().__init__(session=connection._session, **kwargs)
        self.connection = connection

    def request(self, *args, **kwargs):
        if self.connection._is_session_expired():
            self.connection._renew_or_reconnect()

        return super().request(*args, **kwargs)


def renew_session(func):
    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        if (
            not kwargs.pop('skip_expiration_check', False)
            and self._is_session_expired()
        ):
            self._renew_or_reconnect()

        return func(self, *args, **kwargs)

    return wrapper


def log_request(logger):
    def decorator(func):
        @functools.wraps(func)
        def wrapper(self, url, *args, **kwargs):
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug("method = %s url = '%s'", func.__name__.upper(), url)
                logger.debug("headers = %s", self._session.headers)
                logger.debug("headers additional = %s", kwargs.get('headers'))
                logger.debug("body = %s", kwargs.get('json'))
            return func(self, url, *args, **kwargs)

        return wrapper

    return decorator



from typing import Optional

from mstrio.connection import Connection

from ..schedule import Schedule
from .base_subscription import Subscription
from .content import Content


class EmailSubscription(Subscription):
    """Class representation of MicroStrategy Email Subscription object."""

    def __init__(
        self,
        connection: Connection,
        id: Optional[str] = None,
        subscription_id: Optional[str] = None,
        project_id: Optional[str] = None,
        project_name: Optional[str] = None,
    ):
        """Initialize EmailSubscription object, populates it with I-Server data
        if id or subscription_id is passed.
        Specify either `project_id` or `project_name`.
        When `project_id` is provided (not `None`), `project_name` is omitted.

        Args:
            connection (Connection): MicroStrategy connection object returned
                by `connection.Connection()`
            id (str, optional): ID of the subscription to be initialized, only
                id or subscription_id have to be provided at once, if both are
                provided id will take precedence
            subscription_id (str, optional): ID of the subscription to be
                initialized
            project_id (str, optional): Project ID
            project_name (str, optional): Project name
        """
        super().__init__(connection, id, subscription_id, project_id, project_name)

    @classmethod
    def create(
        cls,
        connection: Connection,
        name: str,
        project_id: Optional[str] = None,
        project_name: Optional[str] = None,
        allow_delivery_changes: Optional[bool] = None,
        allow_personalization_changes: Optional[bool] = None,
        allow_unsubscribe: bool = True,
        send_now: Optional[bool] = None,
        owner_id: Optional[str] = None,
        schedules: Optional[str | list[str] | Schedule | list[Schedule]] = None,
        contents: Optional[Content] = None,
        recipients: list[str] | list[dict] = None,
        delivery_expiration_date: Optional[str] = None,
        delivery_expiration_timezone: Optional[str] = None,
        contact_security: Optional[bool] = None,
        space_delimiter: Optional[str] = None,
        email_subject: Optional[str] = None,
        email_message: Optional[str] = None,
        email_send_content_as: str = 'data',
        overwrite_older_version: bool = False,
        filename: Optional[str] = None,
        compress: bool = False,
        zip_filename: Optional[str] = None,
        zip_password_protect: Optional[bool] = None,
        zip_password: Optional[str] = None,
    ):
        """Creates a new email subscription.

        Args:
            connection (Connection): a MicroStrategy connection object
            name (str): name of the subscription,
            project_id (str, optional): project ID,
            project_name (str, optional): project name,
            allow_delivery_changes (bool, optional): whether the recipients can
                change the delivery of the subscription,
            allow_personalization_changes (bool, optional): whether the
                recipients can personalize the subscription,
            allow_unsubscribe (bool, optional): whether the recipients can
                unsubscribe from the subscription,
            send_now (bool, optional): indicates whether to execute the
                subscription immediately,
            owner_id (str, optional): ID of the subscription owner, by default
                logged in user ID,
            schedules (str | list[str] | Schedule | list[Schedule], optional):
                Schedules IDs or Schedule objects,
            contents (Content, optional): The content settings.
            recipients (list[str] | list[dict], optional): list of recipients
                IDs or dicts,
            delivery_expiration_date (str, optional): expiration date of the
                subscription, format should be yyyy - MM - dd,
            delivery_expiration_timezone (str, optional): expiration timezone
                of the subscription, example value 'Europe/London'
            contact_security (bool): whether to use contact security for each
                contact group member
            filename (str, optional): the filename that will be delivered when
                the subscription is executed,
            compress (bool): whether to compress the file
            space_delimiter (str, optional): space delimiter,
            email_subject (str, optional): email subject associated with
                the subscription,
            email_message (str, optional): email body of subscription,
            email_send_content_as (str, enum): [data, data_and_history_list,
                data_and_link_and_history_list, link_and_history_list],
            overwrite_older_version (bool): whether the current subscription
                will overwrite earlier versions of the same report or document
                in the history list,
            zip_filename (str, optional): filename of the compressed content,
            zip_password_protect (bool, optional): whether to password protect
                zip file,
            zip_password (str, optional): optional password for the compressed
                file
        """
        return super()._Subscription__create(
            connection=connection,
            name=name,
            project_id=project_id,
            project_name=project_name,
            allow_delivery_changes=allow_delivery_changes,
            allow_personalization_changes=allow_personalization_changes,
            allow_unsubscribe=allow_unsubscribe,
            send_now=send_now,
            owner_id=owner_id,
            schedules=schedules,
            contents=contents,
            recipients=recipients,
            delivery_mode='EMAIL',
            delivery_expiration_date=delivery_expiration_date,
            delivery_expiration_timezone=delivery_expiration_timezone,
            contact_security=contact_security,
            email_subject=email_subject,
            email_message=email_message,
            filename=filename,
            compress=compress,
            space_delimiter=space_delimiter,
            email_send_content_as=email_send_content_as,
            overwrite_older_version=overwrite_older_version,
            zip_filename=zip_filename,
            zip_password_protect=zip_password_protect,
            zip_password=zip_password,
        )



from typing import List, TYPE_CHECKING, Union, Optional

from mstrio.api import library
from mstrio.project_objects.document import Document, list_documents
from mstrio.project_objects.dossier import list_dossiers
from mstrio.utils.helper import get_valid_project_id
from mstrio.connection import Connection

if TYPE_CHECKING:
    from mstrio.project_objects.dossier import Dossier


class Library:
    def __init__(
        self,
        connection: Connection,
        project_id: Optional[str] = None,
        project_name: Optional[str] = None,
    ):
        self.connection = connection
        ids = self.__get_library_ids()
        self.user_id = connection.user_id

        try:
            project_id = get_valid_project_id(
                connection=connection,
                project_id=project_id,
                project_name=project_name,
                with_fallback=False if project_name else True,
            )
        except ValueError:
            self._documents = None
            self._dossiers = None
            self._contents = None
            return

        self._documents = list_documents(self.connection, project_id=project_id, id=ids)
        self._dossiers = list_dossiers(self.connection, project_id=project_id, id=ids)
        self._contents = self._documents + self._dossiers

    def __get_library_ids(self):
        response = library.get_library(self.connection)
        body = response.json()
        ids = [doc_body['target']['id'] for doc_body in body if 'target' in doc_body]
        return ids

    @property
    def dossiers(self):
        if self.connection.project_id is not None:
            ids = self.__get_library_ids()
            self._dossiers = list_dossiers(self.connection, id=ids)
        return self._dossiers

    @property
    def documents(self):
        if self.connection.project_id is not None:
            ids = self.__get_library_ids()
            self._documents = list_documents(self.connection, id=ids)
        return self._documents

    @property
    def contents(self):
        if self.connection.project_id is not None:
            self._contents = self.dossiers + self.documents
        return self._contents

    def publish(self, contents: Union[List, "Document", "Dossier", str]):
        """Publishes dossier or document to the authenticated user's library.

        contents: dossiers or documents to be published, can be Dossier/Document
            class object or ID
        """
        if not isinstance(contents, list):
            contents = [contents]
        for doc in contents:
            doc_id = doc.id if isinstance(doc, Document) else doc
            body = {'id': doc_id, 'recipients': [{'id': self.user_id}]}
            library.publish_document(self.connection, body=body)

    def unpublish(self, contents: Union[List, "Document", "Dossier", str]):
        """Publishes dossier or document to the authenticated user's library.

        contents: dossiers or documents to be published, can be Dossier/Document
            class object or ID
        """
        if not isinstance(contents, list):
            contents = [contents]
        for doc in contents:
            doc_id = doc.id if isinstance(doc, Document) else doc
            library.unpublish_document_for_user(
                self.connection,
                user_id=self.user_id,
                document_id=doc_id,
            )



from typing import Optional, Type

from mstrio.connection import Connection
from mstrio.types import ObjectTypes
from mstrio.utils.entity import Entity
from mstrio.utils.helper import _prepare_objects as filter_objects
from mstrio.utils.response_processors import drivers, objects
from mstrio.utils.version_helper import class_version_handler


def list_drivers(
    connection: Connection, to_dictionary: bool = False, **filters
) -> list['Driver'] | list[dict]:
    """Get all drivers as list of Driver objects or dictionaries.

    Optionally filter the drivers by specifying filters.

    Args:
        connection: MicroStrategy connection object
        to_dictionary: If True returns a list of Driver dicts,
           otherwise returns a list of Driver objects
       **filters: Available filter parameters:
           ['id', 'name', 'is_enabled', 'is_odbc']
    """
    return Driver.list(connection, to_dictionary, **filters)


@class_version_handler('11.3.0960')
class Driver(Entity):
    """Object representation of Microstrategy Driver

    Attributes:
        connection: A MicroStrategy connection object
        id: Driver's ID
        name: Driver's name
        is_enabled: specifies if a Driver is enabled
        is_odbc: specifies if a Driver is of ODBC type
        description: Object description
        abbreviation: Object abbreviation
        type: Object type
        subtype: Object subtype
        ext_type: Object extended type
        date_created: Creation time, DateTime object
        date_modified: Last modification time, DateTime object
        version: Version ID
        owner: User object that is the owner
        icon_path: Object icon path
        view_media: View media settings
        ancestors: List of ancestor folders
        certified_info: Certification status, time of certification, and
            information about the certifier (currently only for document and
            report)
        acg: Access rights (See EnumDSSXMLAccessRightFlags for possible values)
        acl: Object access control list
    """

    _OBJECT_TYPE = ObjectTypes.DRIVER
    _FROM_DICT_MAP = {
        **Entity._FROM_DICT_MAP,
    }
    _API_GETTERS = {
        (
            'name',
            'description',
            'abbreviation',
            'type',
            'subtype',
            'ext_type',
            'date_created',
            'date_modified',
            'version',
            'owner',
            'icon_path',
            'view_media',
            'ancestors',
            'certified_info',
            'acg',
            'acl',
        ): objects.get_info,
        (
            'id',
            'name',
            'is_enabled',
            'is_odbc',
        ): drivers.get,
    }

    def __init__(
        self,
        connection: Connection,
        id: Optional[str] = None,
        name: Optional[str] = None,
    ):
        """Initialize Driver object by passing ID or name.
        When `id` is provided, `name` is omitted.

        Args:
            connection (Connection): MicroStrategy connection object
            id (str): ID of Driver
            name (str): name of Driver
        """

        if id is None and name is None:
            raise ValueError(
                "Please specify either 'id' or 'name' parameter in the constructor."
            )

        if id is None:
            result: list[dict] = self.list(connection, name=name, to_dictionary=True)

            if result:
                object_data = result[0]
                self._init_variables(**object_data, connection=connection)
            else:
                raise ValueError(f"There is no Driver named: '{name}'.")
        elif id == '':
            raise ValueError("Driver's ID cannot be an empty string.")
        else:
            super().__init__(connection, id)

    def _init_variables(self, **kwargs) -> None:
        super()._init_variables(**kwargs)

        self.is_enabled = kwargs.get('is_enabled')
        self.is_odbc = kwargs.get('is_odbc')

    def alter(
        self,
        is_enabled: Optional[bool] = None,
    ):
        """Update properties of a Driver

        Args:
           is_enabled (bool, optional): specifies if a Driver is enabled
        """

        body = {
            'operationList': [
                {
                    'op': 'replace',
                    'path': '/enabled',
                    'value': is_enabled,
                }
            ]
        }

        response_data = drivers.update(self.connection, id=self.id, body=body)
        self._set_object_attributes(**response_data)

    @classmethod
    def list(
        cls, connection: Connection, to_dictionary: bool = False, **filters
    ) -> list[Type['Driver']] | list[dict]:
        """Get all driver as list of Driver objects or dictionaries.

        Optionally filter the drivers by specifying filters.

        Args:
            connection (Connection): MicroStrategy connection object
            to_dictionary (bool): If True returns a list of Driver dicts,
               otherwise returns a list of Driver objects
           **filters: Available filter parameters:
               ['id', 'name', 'is_enabled', 'is_odbc']
        """

        objects = filter_objects(drivers.get_all(connection), filters)

        if to_dictionary:
            return objects

        return cls.bulk_from_dict(source_list=objects, connection=connection)

    def enable(self):
        """Enable driver."""
        if not self.is_enabled:
            self.alter(is_enabled=True)

    def disable(self):
        """Disable driver."""
        if self.is_enabled:
            self.alter(is_enabled=False)



import getpass
import logging
from enum import auto
from typing import TYPE_CHECKING, Iterable, Optional

import numpy as np
import pandas as pd

from mstrio import config
from mstrio.api import administration, monitors, registrations
from mstrio.connection import Connection
from mstrio.utils.enum_helper import AutoName, AutoUpperName
from mstrio.utils.helper import (
    exception_handler,
    filter_list_of_dicts,
    validate_param_value,
)
from mstrio.utils.version_helper import method_version_handler

from .node import Node, Service, ServiceWithNode

if TYPE_CHECKING:
    from pandas.io.formats.style import Styler

    from mstrio.server.project import Project

logger = logging.getLogger(__name__)


class GroupBy(str, AutoName):
    NODES = auto()
    SERVICES = auto()


class ServiceAction(str, AutoUpperName):
    START = auto()
    STOP = auto()


class Cluster:
    """Manage, list nodes (servers) on a cluster.

    Manage Services on nodes. Manage node settings. Load and Unload
    projects. A "service" is a product developed by MicroStrategy or
    a third-party product distributed by MicroStrategy i.e.
    "MicroStrategy Intelligence Server" or "Apache ZooKeeper".
    """

    def __init__(self, connection: Connection):
        """Initialize Cluster object.

        Args:
            connection: MicroStrategy connection object returned by
                `connection.Connection()`.
        """
        self.connection = connection

    @method_version_handler('11.2.0000')
    def list_nodes(
        self,
        project: "Optional[str | Project]" = None,
        node: Optional[str | Node] = None,
        to_dictionary: bool = False,
    ) -> list[Node | dict]:
        """Return a list of nodes and their properties within the cluster.

        Optionally filter by `project_id` or `node_name`.

        Args:
            project: Project ID or object
            node: Node name or object
        """
        from mstrio.server.project import Project

        project_id = project.id if isinstance(project, Project) else project
        node_name = node.name if isinstance(node, Node) else node
        response = monitors.get_node_info(self.connection, project_id, node_name)
        node_dicts = response.json()['nodes']
        if to_dictionary:
            return node_dicts
        else:
            return [Node.from_dict(node) for node in node_dicts]

    def list_services(self, group_by: GroupBy | str = GroupBy.NODES) -> list:
        """List services in the cluster grouped by nodes or services.

        When `group_by` is set to `nodes` then in the list for each node there
        is given a list of services which are available within. When `group_by`
        is set to `services` then in the list for each services there is given a
        list of nodes on which current service is available.

        Args:
            group_by: determine by what the list will be grouped. Either
                `nodes` or `services`.
        Returns:
            list of nodes with services in case of `group_by` set to `nodes` or
            list of services with nodes in case of `group_by` set to `services`.
        Raise:
            ValueError if `group_by` is neither equal to `nodes` nor `services`.
        """

        if not isinstance(group_by, GroupBy):
            try:
                group_by = GroupBy(group_by)
            except ValueError:
                raise ValueError(
                    "When listing services it is possible to group them only by "
                    "`nodes` or `services`. See the GroupBy enum."
                )
        if group_by == GroupBy.NODES:
            nodes = registrations.get_nodes(connection=self.connection).json()
            return [
                {
                    'node': Node.from_dict(node['node']),
                    'services': [
                        Service.from_dict(service) for service in node['services']
                    ],
                }
                for node in nodes
            ]
        else:  # GroupBy.SERVICES
            services = registrations.get_services(connection=self.connection).json()
            return [
                {
                    'service': service['service'],
                    'nodes': [
                        ServiceWithNode.from_dict(node) for node in service['nodes']
                    ],
                }
                for service in services
            ]

    def nodes_topology(self, styled=False) -> "pd.DataFrame | Styler":
        """Return cluster node topology as Pandas DataFrame or Styler object.

        DataFrame columns:
            - displayName -> it shows the name with which service is displayed
                in Workstation
            - id -> name of the service within nodes
            - for each node in the cluster there is given a column which name is
                name of a node and its content is information with the status of
                the given service in this node. Available values of status are:
                'Running' (green color), 'Stopped' (red color), 'Not Available
                (service unavailable on the node)

        Returns:
            Pandas DataFrame or Styler object if styled is True.
        """
        nodes = self.list_services(group_by=GroupBy.NODES)
        metadata = registrations.get_services_metadata(
            connection=self.connection
        ).json()
        service_type = pd.DataFrame(metadata["serviceTypes"])[['displayName', 'name']]

        for node in nodes:
            node_name = node['node'].name
            if node['services']:
                node_service = pd.DataFrame(
                    [service.to_dict() for service in node['services']]
                )[['id', 'status']]
                node_service['status'] = node_service['status'].map(
                    lambda x: 'Running' if x == "PASSING" else 'Stopped'
                )
                tp = service_type.join(
                    node_service.set_index('id'),
                    on='name',
                    how='inner',
                )[['status']]
                # add column with statuses in current node
                service_type[node_name] = tp['status']
            else:
                # add an empty column in case of no services for the given node
                service_type[node_name] = np.nan

            service_type[node_name] = service_type[node_name].fillna('Not Available')

        service_type = service_type.rename(columns={'name': 'id'}, inplace=False)

        if styled:
            return service_type.style.applymap(Cluster._show_color)
        else:
            return service_type

    def services_topology(self, styled=False) -> "pd.DataFrame | Styler":
        """Return cluster service topology as Pandas DataFrame or Styler object.

        DataFrame columns:
            - service -> name of the service
            - node -> name on which service appears (each service can appear in
                more than one service and then name of the service is provided
                only once for the group of nodes in which it appears)
            - status -> status which a given service has on a given node.
                Available values of statuses are: 'Running' (green color),
                'Stopped' (red color), 'Not Available (service unavailable on
                the node)

        Returns:
            Pandas DataFrame or Styler object if styled is True.
        """
        services = self.list_services(group_by=GroupBy.SERVICES)
        services_topology = [
            {
                "service": s['service'],
                "nodes": [{"node": n.node, "status": n.status} for n in s['nodes']],
            }
            for s in services
        ]

        tmp = []
        for s in services_topology:
            for n in s['nodes']:
                n['service'] = s['service']
                tmp.append(n)

        tmp_df = pd.DataFrame(tmp)[['service', 'node', 'status']]
        tmp_df['status'] = tmp_df['status'].map(
            lambda x: 'Running' if x == "PASSING" else 'Stopped'
        )
        tmp_df = tmp_df.set_index('service', append=True).swaplevel(0, 1)

        if styled:
            return tmp_df.style.applymap(Cluster._show_color, subset='status')
        else:
            return tmp_df

    @method_version_handler('11.2.0000')
    def add_node(self, node: str | Node) -> None:
        """Add server (node) to the cluster.

        Args:
            node: name or object of node to be added
        """
        name = node.name if isinstance(node, Node) else node
        monitors.add_node(self.connection, name)
        if config.verbose:
            logger.info(f'{name} added to cluster.')

    @method_version_handler('11.2.0000')
    def remove_node(self, node: str | Node) -> None:
        """Remove server (node) from the cluster.

        Args:
            node: name or object of node to be removed
        """
        name = node.name if isinstance(node, Node) else node
        monitors.remove_node(self.connection, name)
        if config.verbose:
            logger.info(f'{name} removed from cluster.')

    def set_primary_node(self, node: str | Node) -> None:
        """Set default/primary server (node) for the cluster.

        Args:
            name: name or object of the node which will be set as default
                for this cluster
        """
        name = node.name if isinstance(node, Node) else node
        body = {"defaultHostname": name}
        res = administration.update_iserver_configuration_settings(
            connection=self.connection, body=body
        )
        if res.ok and config.verbose:
            logger.info(f'Primary node of the cluster was set to {name}.')

    def check_dependency(self, service: str) -> list[str]:
        """Check all dependencies for the given service.

        Raises:
            ValueError: If incorrect service name is provided.
        """
        if not hasattr(self, '_metadata'):
            self._metadata = registrations.get_services_metadata(
                connection=self.connection
            ).json()
        service_id_map = {
            service['typeId']: service['name']
            for service in self._metadata['serviceTypes']
        }
        available_services = list(service_id_map.values())

        def get_dependencies_recursively(service_name: str):
            service = filter_list_of_dicts(
                self._metadata['serviceTypes'], name=service_name
            )
            if service:
                service = service[0]
                dependencies_set = {
                    name
                    for id, name in service_id_map.items()
                    if id in service['dependsOn']
                }

                if dependencies_set:
                    for srv in dependencies_set:
                        dependencies_set.update(get_dependencies_recursively(srv))
                    return dependencies_set
                else:
                    return {}
            else:
                raise ValueError(
                    f"Service {service_name} is incorrect. Please choose one of: "
                    f"{available_services}"
                )

        return list(get_dependencies_recursively(service))

    def start(
        self,
        service: str,
        nodes: list[str],
        login: Optional[str] = None,
        passwd: Optional[str] = None,
    ):
        """Start up a service on selected nodes.

        Args:
            service: name of the service which will be started
            nodes: list of names of nodes on which service will be started
            login: login for SSH operation. If not provided, the user will be
                prompted.
            passwd: password for SSH operation. If not provided, the user will
                be prompted.
        Raises:
            ValueError: If incorrect node/service name is provided
        """
        self._control_service(ServiceAction.START, service, nodes, login, passwd)

    def stop(
        self,
        service: str,
        nodes: list[str],
        login: Optional[str] = None,
        passwd: Optional[str] = None,
        force: bool = False,
    ):
        """Stop a service on selected nodes. Provided service and node names
        are checked for correctness.

        Args:
            service: name of the service which will be started
            nodes: list of names of nodes on which service will be started
            login: login for SSH operation. If not provided, the user will be
                prompted.
            passwd: password for SSH operation. If not provided, the user will
                be prompted.
            force: if True, no additional prompt will be shown before
        Raises:
            ValueError: If incorrect node/service name is provided
        """
        self._control_service(ServiceAction.STOP, service, nodes, login, passwd, force)

    def _control_service(
        self,
        action: ServiceAction,
        service: str,
        nodes: list[str],
        login: Optional[str] = None,
        passwd: Optional[str] = None,
        force: bool = False,
    ):
        # validate inputs
        self._check_nodes(nodes)
        service_list = self.list_services(group_by=GroupBy.SERVICES)
        self._check_service(service, service_list)

        # ask for confirmation when stopping MicroStrategy-Intelligence-Server
        if action == ServiceAction.STOP:
            if not force and service == 'MicroStrategy-Intelligence-Server':
                msg = (
                    "Stopping the Intelligence Server can affect all the users' "
                    "sessions, including this current session."
                )
                logger.info(msg)
                if input("Are you sure you want to proceed? [Y/N]:") != 'Y':
                    return
        if action == ServiceAction.START:
            self._check_dependencies(service, service_list)
        # get credentials for operation on service
        if not login or not passwd:
            print("Provide credentials for SSH operation.")
        if not login:
            login = input("username: ")
        if not passwd:
            passwd = getpass.getpass("password: ")
        wrong = []
        good = []
        # try to start each node; in case of error, execution is not terminated
        for node_name in nodes:
            info = Cluster._get_node_info(node_name, service, service_list)
            if not info:
                wrong.append(node_name)
                continue
            result = registrations.start_stop_service(
                connection=self.connection,
                login=login,
                password=passwd,
                name=service,
                id=info['id'],
                address=info['address'],
                action=action.value,
            ).ok
            if not result:
                wrong.append(node_name)
            else:
                good.append(node_name)
        self._show_start_stop_msg(service, wrong, good, action)

    def list_node_settings(self, node: str | Node) -> dict:
        """List server (nodes) settings.

        Args:
            node: name or object of node which settings will be listed

        Returns:
            dictionary with the settings of node returned from I-Server.
        """
        node_name = node.name if isinstance(node, Node) else node
        response = administration.get_iserver_node_settings(self.connection, node_name)
        return response.json()

    def update_node_settings(
        self,
        node: str | Node,
        load_balance_factor: int,
        initial_pool_size: int,
        max_pool_size: int,
    ) -> None:
        """Update I-Server configuration settings for a given server node
        within a cluster.

        Args:
            load_balance_factor: This setting becomes relevant in an environment
                that has a MicroStrategy Intelligence Server cluster. By
                default, the load balance factor is 1. The value can be
                increased on more powerful servers in a cluster to provide an
                appropriate balance. A larger load balance factor means the
                current server consumes a greater load in the server cluster in
                which it resides.
            initial_pool_size: Initial number of connections available.
            max_pool_size: Maximum number of connections available.
        """
        validate_param_value("load_balance_factor", load_balance_factor, int, 100, 1)
        validate_param_value("initial_pool_size", initial_pool_size, int, 1024, 1)
        validate_param_value("max_pool_size", max_pool_size, int, 1024, 1)

        node_name = node.name if isinstance(node, Node) else node
        body = {
            "loadBalanceFactor": load_balance_factor,
            "initialPoolSize": initial_pool_size,
            "maxPoolSize": max_pool_size,
        }
        response = administration.update_iserver_node_settings(
            self.connection, body, node_name
        )
        if config.verbose and response.ok:
            logger.info(f'Intelligence Server configuration updated for {node_name}')

    def reset_node_settings(self, node: str | Node) -> None:
        """Remove I-Server configuration settings for given node within a
        cluster. Default values will be applied after execution of this method.

        Args:
            node: name of the node for which default settings will be applied.
        """
        node_name = node.name if isinstance(node, Node) else node
        administration.delete_iserver_node_settings(self.connection, node_name)

    def list_projects(
        self, to_dictionary: bool = False, limit: Optional[int] = None, **filters
    ) -> list["Project"]:
        """Return list of project objects or if `to_dictionary=True`
        project dicts. Optionally filter the Projects by specifying the
        `filters` keyword arguments.

        Args:
            to_dictionary: If True returns list of project dicts
            limit: limit the number of elements returned. If `None`, all objects
                are returned.
            **filters: Available filter parameters: ['name', 'id',
                'description', 'date_created', 'date_modified', 'owner']
        """
        from mstrio.server.environment import Environment

        env = Environment(connection=self.connection)
        return env.list_projects(to_dictionary=to_dictionary, limit=limit, **filters)

    def load_project(
        self, project: "str | Project", on_nodes: Optional[str | list[str]] = None
    ) -> None:
        """Request to load the project onto the chosen cluster nodes. If
        nodes are not specified, the project will be loaded on all nodes.

        Args:
            project: name or object of project which will be loaded
            on_nodes: name of node or nodes, if not passed, project will be
                loaded on all of the nodes
        """

        from mstrio.server.project import Project

        project_name = project.name if isinstance(project, Project) else project
        project = Project._list_projects(self.connection, name=project_name)[0]
        project.load(on_nodes=on_nodes)

    def unload_project(
        self, project: "str | Project", on_nodes: Optional[str | list[str]] = None
    ) -> None:
        """Request to unload the project from the chosen cluster nodes. If
        nodes are not specified, the project will be unloaded on all nodes.
        The unload action cannot be performed until all jobs and connections
        for project are completed. Once these processes have finished,
        pending project will be automatically unloaded.

        Args:
            project: name or object of project which will be unloaded
            on_nodes: name of node or nodes, if not passed, project will be
                unloaded on all of the nodes
        """
        from mstrio.server.project import Project

        project_name = project.name if isinstance(project, Project) else project
        project = Project._list_projects(self.connection, name=project_name)[0]
        project.unload(on_nodes=on_nodes)

    def _show_start_stop_msg(
        self,
        service_name: str,
        wrong: list[str],
        good: list[str],
        action: ServiceAction,
    ):
        """Prepare message to show after action of stopping or starting a
        service with the information on which node the actions were good (done
        correctly) and on which were wrong (response status was not ok).
        """
        if len(wrong) > 0:
            action_msg = 'started' if action == ServiceAction.START else 'stopped'
            nodes_msg = ', '.join(wrong)
            logger.warning(
                f'Service {service_name} was not {action_msg} for node(s) {nodes_msg}.'
            )
        if len(good) > 0:
            action_msg = 'start' if action == ServiceAction.START else 'stop'
            nodes_msg = ','.join(good)
            logger.info(
                f'Request to {action_msg} {service_name} was sent for node(s) '
                f'{nodes_msg}.'
            )

    def _check_service(
        self, service_name: str, service_list: list[ServiceWithNode]
    ) -> None:
        """Checks if the name of the given service is one of the names of
        existing services.

        Raises:
            ValueError: If name is incorrect.
        """
        valid_service_names = [service['service'] for service in service_list]

        if service_name not in valid_service_names:
            raise ValueError(
                f"Service {service_name} is incorrect. Please choose one of: "
                f"{valid_service_names}"
            )

    def _check_dependencies(
        self, service_name: str, service_list: list[ServiceWithNode]
    ):
        """Check if service depends on other services.

        Warns if any one of the dependencies is not running.
        """
        dependencies = self.check_dependency(service_name)
        dependencies = [
            service
            for service in dependencies
            if not Cluster._check_service_running(service_name, service_list)
        ]

        if dependencies:
            exception_handler(
                f"Service {service_name} depends on services {dependencies} to run "
                "correctly.",
                Warning,
            )

    def _check_nodes(self, node_names: Iterable[str]) -> None:
        """Checks if the names of the given nodes are within names of existing
        nodes. Checks if the node is running.

        Raises:
            ValueError: If name is incorrect.
        """
        all_nodes = self.list_nodes(to_dictionary=True)
        all_node_names = [node['name'] for node in all_nodes]
        active_nodes = [
            node['name']
            for node in filter(lambda node: node['status'] == 'running', all_nodes)
        ]

        for node_name in node_names:
            if node_name not in all_node_names:
                raise ValueError(
                    f"Node '{node_name}' is incorrect. Please choose one of: "
                    f"{all_node_names}"
                )
            elif node_name not in active_nodes:
                raise ValueError(
                    f"Node '{node_name}' is stopped. Try starting it first."
                )
            else:
                continue

    @staticmethod
    def _check_service_running(
        service_name: str,
        service_list: list[ServiceWithNode],
        node_name: Optional[str] = None,
    ) -> bool:
        """Return True if service is running on any node available.

        If `node_name` is provided, the service status will be given for
        the selected node.
        """
        nodes = [
            service['nodes']
            for service in service_list
            if service['service'] == service_name
        ][0]

        if node_name:
            node = [node for node in nodes if node.node == node_name]
            if node.status == 'PASSING':
                return True
            else:
                return False
        else:
            return bool([True for node in nodes if node.status == 'PASSING'])

    @staticmethod
    def _get_node_info(
        node_name: str, service_name: str, service_list: list[ServiceWithNode]
    ) -> Optional[dict]:
        nodes = [
            service['nodes']
            for service in service_list
            if service['service'] == service_name
        ][0]
        node = [node for node in nodes if node.node == node_name]

        if not node:
            exception_handler(
                f"Service {service_name} is not available on {node_name}",
                exception_type=Warning,
            )
            return None
        else:
            node = node[0]

        if node.service_control is False:
            exception_handler(
                f"Service {service_name} cannot be controlled on {node_name}",
                exception_type=Warning,
            )
            return None
        else:
            return node.to_dict()

    @staticmethod
    def _show_color(val: str) -> str:
        """Prepare color in which values of columns in DataFrames for
        topologies are shown."""
        match val:
            case 'Running':
                color = 'Green'
            case 'Stopped':
                color = '#b22222'
            case _:
                color = 'Black'

        return 'color: %s' % color

    @property
    def default_node(self) -> Optional[str]:
        """Return name of default node for this cluster."""

        for n in self.list_nodes(to_dictionary=True):
            if n.get('default', 'False') is True:
                return n.get('name')
        return None



from dataclasses import dataclass
from enum import auto
import logging
from typing import Optional

from mstrio.modeling import SchemaObjectReference
from mstrio.utils.enum_helper import AutoName
from mstrio.utils.helper import Dictable, snake_to_camel
from mstrio.utils.wip import module_wip, WipLevels

logger = logging.getLogger(__name__)

module_wip(globals(), level=WipLevels.WARNING)


class PropertyLocation(AutoName):
    REPORT_TARGET = auto()
    REPORT = auto()
    TEMPLATE_TARGET = auto()
    TEMPLATE = auto()
    OBJECT = auto()
    PROJECT = auto()
    DB_ROLE = auto()
    DBMS = auto()
    DEFAULT = auto()


@dataclass
class AttributeJoinType(Dictable):
    _FROM_DICT_MAP = {
        'attribute': SchemaObjectReference.from_dict,
        'resolved_location': PropertyLocation,
    }
    attribute: SchemaObjectReference
    no_participation_in_preserve_row: bool
    resolved_location: PropertyLocation


class JoinType(AutoName):
    INNER = auto()
    OUTER = auto()


@dataclass
class MetricJoinType(Dictable):
    _FROM_DICT_MAP = {
        'metric': SchemaObjectReference.from_dict,
        'join_type': JoinType,
        'resolved_location': PropertyLocation,
    }
    metric: SchemaObjectReference
    join_type: JoinType
    resolved_location: PropertyLocation


class VLDBPropertyType(AutoName):
    """Enumeration constant indicating type of value of a VLDB property value"""

    INT32 = auto()
    INT64 = auto()
    DATE = auto()
    TIME = auto()
    BOOLEAN = auto()
    STRING = auto()
    DOUBLE = auto()


@dataclass
class VLDBProperty(Dictable):
    _FROM_DICT_MAP = {
        'type': VLDBPropertyType,
        'resolved_location': PropertyLocation,
    }
    name: str
    value: str
    type: VLDBPropertyType
    resolved_location: PropertyLocation


@dataclass
class AdvancedProperties(Dictable):
    _KEEP_CAMEL_CASE = [
        'vldbProperties',
        'metricJoinTypes',
        'attributeJoinTypes',
        'vldb_properties',
        'metric_join_types',
        'attribute_join_types',
    ]

    @staticmethod
    def _parse_metric(source, connection):
        return {
            key: MetricJoinType.from_dict(value, connection)
            for key, value in source.items()
        }

    @staticmethod
    def _parse_attribute(source, connection):
        return {
            key: AttributeJoinType.from_dict(value, connection)
            for key, value in source.items()
        }

    @staticmethod
    def _parse_vldb(source, connection):
        return {
            key: VLDBProperty.from_dict(value, connection)
            for key, value in source.items()
        }

    _FROM_DICT_MAP = {
        'vldb_properties': _parse_vldb,
        'metric_join_types': _parse_metric,
        'attribute_join_types': _parse_attribute,
    }
    vldb_properties: Optional[dict] = None
    metric_join_types: Optional[dict] = None
    attribute_join_types: Optional[dict] = None

    def to_dict(self, camel_case: bool = True) -> dict:
        result = {
            'vldb_properties': {
                key: value.to_dict(camel_case)
                for key, value in self.vldb_properties.items()
            }
            if self.vldb_properties
            else None,
            'metric_join_types': {
                key: value.to_dict(camel_case)
                for key, value in self.metric_join_types.items()
            }
            if self.metric_join_types
            else None,
            'attribute_join_types': {
                key: value.to_dict(camel_case)
                for key, value in self.attribute_join_types.items()
            }
            if self.attribute_join_types
            else None,
        }

        return (
            snake_to_camel(result, whitelist=self._KEEP_CAMEL_CASE)
            if camel_case
            else result
        )



from typing import Optional, TYPE_CHECKING

from mstrio.utils.api_helpers import changeset_manager, unpack_information
from mstrio.utils.error_handlers import ErrorHandler

if TYPE_CHECKING:
    from mstrio.connection import Connection


@ErrorHandler(err_msg='Error listing Hierarchies.')
def get_user_hierarchies(
    connection: "Connection",
    project_id: Optional[str] = None,
    changeset_id: Optional[str] = None,
    limit: int = 1000,
    offset: int = 0,
    error_msg: Optional[str] = None,
):
    """Get a list of all user hierarchies.
    The project ID is required to return all user hierarchy definitions
    in metadata. The changeset ID is required to return all user hierarchy
    definitions within a specific changeset. To execute the request,
    provide either the project ID or changeset ID.
    If you provide both, only the changeset ID is used.

    Args:
        connection: MicroStrategy REST API connection object
        project_id (str, optional): Project ID
        changeset_id (str, optional): Changeset ID
        limit (int, optional): Maximum number of items returned for a single
            request. Used to control paging behavior. Use -1 for no limit.
            Default is 1000.
        offset (int, optional): Starting point within the collection of returned
            results. Used to control paging behavior. Default is 0.
        error_msg (str, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. Expected status is 200.
    """
    if project_id is None:
        connection._validate_project_selected()
        project_id = connection.project_id

    return connection.get(
        url=f"{connection.base_url}/api/model/hierarchies",
        headers={
            'X-MSTR-ProjectID': project_id,
            'X-MSTR-MS-Changeset': changeset_id,
        },
        params={
            'limit': limit,
            'offset': offset,
        },
    )


@unpack_information
@ErrorHandler(err_msg='Error creating the user hierarchy.')
def create_user_hierarchy(
    connection: "Connection", body: dict, error_msg: Optional[str] = None
):
    """Creates a new user hierarchy in the changeset,
    based on the definition provided in request body.

    Args:
        connection: MicroStrategy REST API connection object
        body: User hierarchy creation body
        error_msg (str, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. Expected status is 201.
    """
    with changeset_manager(connection) as changeset_id:
        return connection.post(
            url=f"{connection.base_url}/api/model/hierarchies",
            headers={'X-MSTR-MS-Changeset': changeset_id},
            json=body,
        )


@unpack_information
@ErrorHandler(err_msg='Error getting the user hierarchy with ID: {id}.')
def get_user_hierarchy(
    connection: "Connection",
    id: str,
    project_id: Optional[str] = None,
    changeset_id: Optional[str] = None,
    error_msg: Optional[str] = None,
):
    """Get the definition of a user hierarchy.
    The project ID is required to return a user hierarchy's definition
    in metadata. The changeset ID is required to return a user hierarchy's
    definition within a specific changeset. To execute the request, provide
    either the project ID or changeset ID.
    If both are provided, only the changeset ID is used.

    Args:
        connection: MicroStrategy REST API connection object
        id (str): Hierarchy ID. The ID can be:
            - the object ID used in the metadata.
            - the object ID used in the changeset, but not yet committed
            to metadata.
        project_id (str, optional): Project ID
        changeset_id (str, optional): Changeset ID
        error_msg (str, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. Expected status is 200.
    """
    if project_id is None:
        connection._validate_project_selected()
        project_id = connection.project_id

    return connection.get(
        url=f"{connection.base_url}/api/model/hierarchies/{id}",
        headers={'X-MSTR-ProjectID': project_id, 'X-MSTR-MS-Changeset': changeset_id},
    )


@unpack_information
@ErrorHandler(err_msg='Error updating the user hierarchy with ID: {id}.')
def update_user_hierarchy(
    connection: "Connection", id: str, body: dict, error_msg: Optional[str] = None
):
    """Updates a specific user hierarchy in the changeset,
    based on the definition provided in the request body.

    Args:
        connection: MicroStrategy REST API connection object
        id (str): Hierarchy ID. The ID can be:
            - the object ID used in the metadata.
            - the object ID used in the changeset, but not yet committed
            to metadata.
        body: User hierarchy update info
        error_msg (str, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. Expected status is 200.
    """
    with changeset_manager(connection) as changeset_id:
        return connection.patch(
            url=f"{connection.base_url}/api/model/hierarchies/{id}",
            headers={'X-MSTR-MS-Changeset': changeset_id},
            json=body,
        )


@ErrorHandler(err_msg='Error deleting the user hierarchy with ID: {id}.')
def delete_user_hierarchy(
    connection: "Connection", id: str, error_msg: Optional[str] = None
):
    """Delete a specific user hierarchy

    Args:
        connection: MicroStrategy REST API connection object
        id (str): Hierarchy ID. The ID can be:
            - the object ID used in the metadata.
            - the object ID used in the changeset, but not yet committed
            to metadata.
        error_msg (str, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. Expected status is 204.
    """
    with changeset_manager(connection) as changeset_id:
        return connection.delete(
            url=f"{connection.base_url}/api/model/hierarchies/{id}",
            headers={'X-MSTR-MS-Changeset': changeset_id},
        )



# NOSONAR
from collections import defaultdict
from enum import auto
import logging
from typing import Iterable, Optional, Union, List, TYPE_CHECKING

from mstrio import config
from mstrio.api import contact_groups
from mstrio.users_and_groups.user import User
from mstrio.utils.entity import auto_match_args_entity, DeleteMixin, EntityBase
from mstrio.utils.enum_helper import AutoName
from mstrio.utils.helper import (
    Dictable,
    fetch_objects_async,
    get_args_from_func,
    get_default_args_from_func,
    get_objects_id,
)
from mstrio.utils.version_helper import class_version_handler, method_version_handler

if TYPE_CHECKING:
    from mstrio.connection import Connection
    from mstrio.users_and_groups.contact import Contact

logger = logging.getLogger(__name__)


class ContactGroupMemberType(AutoName):
    CONTACT = auto()
    CONTACT_GROUP = auto()
    USER = auto()


class ContactGroupMember(Dictable):
    """ContactGroupMember class, representing either Contact or ContactGroup

    Attributes:
        name: member's name
        type: type of a member, instance of ContactGroupMemberType
        id: id of member, optional
        description: member's description, optional
        enabled: specifies if a member is enabled
    """

    _FROM_DICT_MAP = {"type": ContactGroupMemberType}

    def __init__(
        self,
        name: str,
        type: Union[str, ContactGroupMemberType],
        id: Optional[str] = None,
        description: Optional[str] = None,
        enabled: bool = True,
    ):
        self.name = name
        self.type = ContactGroupMemberType(type) if isinstance(type, str) else type
        self.id = id
        self.description = description
        self.enabled = enabled

    def __repr__(self) -> str:
        param_dict = auto_match_args_entity(
            self.__init__, self, exclude=['self'], include_defaults=False
        )

        params = [
            f"{param}={self.type}" if param == 'type' else f'{param}={repr(value)}'
            for param, value in param_dict.items()
        ]
        formatted_params = ', '.join(params)

        return f'ContactGroupMember({formatted_params})'

    @classmethod
    def from_contact_or_contact_group(
        cls, obj: Union['Contact', 'ContactGroup']
    ) -> 'ContactGroupMember':
        """Initialize instance of class ContactGroupMember

        Args:
            obj: object to used as base for initializing instance of
            ContactGroupMember

        Returns:
            ContactGroupMember object
        """
        from mstrio.users_and_groups.contact import Contact

        if isinstance(obj, Contact):
            return cls(id=obj.id, name=obj.name, type=ContactGroupMemberType.CONTACT)

        if isinstance(obj, ContactGroup):
            return cls(
                id=obj.id, name=obj.name, type=ContactGroupMemberType.CONTACT_GROUP
            )


@method_version_handler('11.3.0200')
def list_contact_groups(
    connection: "Connection",
    to_dictionary: bool = False,
    limit: Optional[int] = None,
    **filters,
) -> Union[List["ContactGroup"], List[dict]]:
    """Get all contact groups as list of ContactGroup objects or
    dictionaries.

    Optionally filter the contact groups by specifying filters.

    Args:
        connection(object): MicroStrategy connection object
        to_dictionary: If True returns a list of contact group dicts,
            otherwise returns a list of contact group objects
        limit: limit the number of elements returned. If `None` (default), all
            objects are returned.
        **filters: Available filter parameters: ['id', 'name', 'description',
            'enabled', 'linked_user']
    """
    return ContactGroup._list_contact_groups(
        connection=connection, to_dictionary=to_dictionary, limit=limit, **filters
    )


@class_version_handler('11.3.0200')
class ContactGroup(EntityBase, DeleteMixin):
    """Object representation of Microstrategy Contact Group object

    Attributes:
        name: contact group's name
        id: contact group's id
        description: contact group's description
        enabled: specifies if a contact group is enabled
        linked_user: user linked to contact group, instance of User
        members: list of contact group's members, instances of
            ContactGroupMember
        memberships: list of Contact Groups that the Contact Group belongs to
        connection: instance of Connection class, represents connection
                    to MicroStrategy Intelligence Server
    """

    _FROM_DICT_MAP = {
        **EntityBase._FROM_DICT_MAP,
        'linked_user': User.from_dict,
        'members': [ContactGroupMember.from_dict],
    }
    _API_GETTERS = {
        (
            'id',
            'name',
            'description',
            'enabled',
            'linked_user',
            'members',
            'memberships',
        ): contact_groups.get_contact_group
    }
    _API_DELETE = staticmethod(contact_groups.delete_contact_group)
    _API_PATCH = {
        ('name', 'description', 'enabled', 'linked_user', 'members'): (
            contact_groups.update_contact_group,
            'put',
        )
    }
    _PATCH_PATH_TYPES = {
        'name': str,
        'description': str,
        'enabled': bool,
        'linked_user': dict,
        'members': list,
        'memberships': list,
    }

    def __init__(
        self,
        connection: 'Connection',
        id: Optional[str] = None,
        name: Optional[str] = None,
    ):
        """Initialize Contact Group object by passing id or name.
        When `id` is provided, name is omitted.

        Args:
            connection: MicroStrategy connection object
            id: ID of Contact
            name: name of Contact Group
        """

        if id is None and name is None:
            raise ValueError(
                "Please specify either 'id' or 'name' parameter in the constructor."
            )

        if id is None:
            result = ContactGroup._list_contact_groups(
                connection=connection,
                name=name,
                to_dictionary=True,
            )
            if result:
                object_data, object_data['connection'] = result[0], connection
                self._init_variables(**object_data)
            else:
                raise ValueError(f"There is no Contact Group named: '{name}'")
        else:
            super().__init__(connection, id)

    def _init_variables(self, **kwargs) -> None:
        super()._init_variables(**kwargs)

        self.description = kwargs.get('description')
        self.enabled = kwargs.get('enabled')

        linked_user = kwargs.get("linked_user")
        self.linked_user = (
            User.from_dict(linked_user, self.connection) if linked_user else None
        )

        members = kwargs.get('members')
        self.members = (
            [ContactGroupMember.from_dict(member) for member in members]
            if members
            else None
        )

        memberships = kwargs.get('memberships')
        self._memberships = (
            [self.from_dict(m, self.connection) for m in memberships]
            if memberships
            else None
        )

    @classmethod
    def create(
        cls,
        connection: "Connection",
        name: str,
        linked_user: Union[str, User],
        members: List[Union[dict, ContactGroupMember]],
        description: Optional[str] = None,
        enabled: bool = True,
    ) -> 'ContactGroup':
        """Create a new contact group.

        Args:
            connection: MicroStrategy connection object
                returned by `connection.Connection()`
            name: contact group name
            linked_user: user linked to contact
            members: list of members
            description: description of contact
            enabled: specifies if contact should be enabled

        Returns:
            `ContactGroup` object
        """
        members = [
            m.to_dict() if isinstance(m, ContactGroupMember) else m for m in members
        ]
        linked_user = get_objects_id(linked_user, User)
        body = {
            'name': name,
            'description': description,
            'enabled': enabled,
            'linkedUser': {'id': linked_user},
            'members': members,
        }
        res = contact_groups.create_contact_group(connection, body).json()
        if config.verbose:
            logger.info(
                f"Successfully created contact group named: '{res.get('name')}' "
                f"with ID: '{res.get('id')}'"
            )
        return cls.from_dict(res, connection)

    def alter(
        self,
        name: Optional[str] = None,
        description: Optional[str] = None,
        enabled: Optional[bool] = None,
        linked_user: Optional[Union['User', str]] = None,
        members: Optional[Iterable[Union['ContactGroupMember', dict]]] = None,
    ):
        """Update properties of a contact group

        Args:
            name: name of a contact
            description: description of a contact
            enabled: specifies if a contact is enabled
            linked_user: an object of class User linked to the contact
            members: list of contact group members, instances of
               `ContactGroupMember`
        """

        linked_user = {'id': get_objects_id(linked_user, User)} if linked_user else None

        func = self.alter
        args = get_args_from_func(func)
        defaults = get_default_args_from_func(func)
        defaults_dict = dict(zip(args[-len(defaults) :], defaults)) if defaults else {}
        local = locals()

        properties = defaultdict(dict)

        for property_key in defaults_dict:
            if local[property_key] is not None:
                properties[property_key] = local[property_key]

        self._alter_properties(**properties)

    @classmethod
    def _list_contact_groups(
        cls,
        connection: "Connection",
        to_dictionary: bool = False,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        **filters,
    ) -> Union[List["ContactGroup"], List[dict]]:
        objects = fetch_objects_async(
            connection=connection,
            api=contact_groups.get_contact_groups,
            async_api=contact_groups.get_contact_groups_async,
            limit=limit,
            offset=offset,
            chunk_size=1000,
            filters=filters,
            dict_unpack_value='contactGroups',
        )

        if to_dictionary:
            return objects
        return [
            ContactGroup.from_dict(source=obj, connection=connection) for obj in objects
        ]

    def _set_object_attributes(self, **kwargs) -> None:
        super()._set_object_attributes(**kwargs)
        memberships = kwargs.get("memberships")

        memberships_objs = (
            [self.from_dict(m, self.connection) for m in memberships]
            if memberships
            else []
        )
        setattr(self, '_memberships', memberships_objs)

    def add_members(
        self, members: Iterable[Union['ContactGroupMember', 'Contact', 'ContactGroup']]
    ):
        """Add member

        Args:
           members: list of members to add to contact group
        """
        members_ids = [member.id for member in self.members]
        new_members = [
            ContactGroupMember.from_contact_or_contact_group(obj)
            if not isinstance(obj, ContactGroupMember)
            else obj
            for obj in members
            if obj.id not in members_ids
        ]

        self.alter(members=new_members + self.members)

    def remove_members(
        self, members: Iterable[Union['ContactGroupMember', 'Contact', 'ContactGroup']]
    ):
        """Remove member

        Args:
           members: list of members to remove from contact group
        """
        ids_to_remove = [member.id for member in members]
        new_member_list = [
            member for member in self.members if member.id not in ids_to_remove
        ]

        self.alter(members=new_member_list)

    @property
    def memberships(self):
        return self._memberships



import logging
from dataclasses import dataclass, field
from enum import auto
from typing import Optional, Type

from mstrio.connection import Connection
from mstrio.modeling import ObjectSubType
from mstrio.utils.enum_helper import AutoName
from mstrio.utils.exceptions import NotSupportedError
from mstrio.utils.helper import Dictable
from mstrio.utils.wip import module_wip, WipLevels

logger = logging.getLogger(__name__)

module_wip(globals(), level=WipLevels.WARNING)


class TemplateUnitType(AutoName):
    ATTRIBUTE = auto()
    DIMENSION = auto()
    METRICS = auto()
    CUSTOM_GROUP = auto()
    CONSOLIDATION = auto()
    PROMPT = auto()
    RAW_UNIT = auto()


@dataclass
class TemplateUnit(Dictable):
    _TYPE = None

    @staticmethod
    def dispatch(
        source, connection: Optional[Connection] = None
    ) -> Type['TemplateUnit']:
        data = source.copy()
        unit_type = TemplateUnitType(data.pop('type'))

        if unit_type == TemplateUnitType.ATTRIBUTE:
            cls = AttributeTemplateUnit
        elif unit_type == TemplateUnitType.METRICS:
            cls = MetricTemplateUnit
        else:
            raise NotSupportedError(
                f"Unit type: '{unit_type}' is not supported for Template class."
            )

        return cls.from_dict(data, connection)

    def to_dict(self, camel_case: bool = True) -> dict:
        return {
            **super().to_dict(camel_case),
            'type': self._TYPE.value,
        }


@dataclass
class AttributeFormReference(Dictable):
    id: str
    name: str


@dataclass
class AttributeTemplateUnit(TemplateUnit):
    _TYPE = TemplateUnitType.ATTRIBUTE
    _FROM_DICT_MAP = {
        'forms': [AttributeFormReference.from_dict],
    }
    id: str
    name: Optional[str] = None
    alias: Optional[str] = None
    forms: Optional[list[AttributeFormReference]] = None


@dataclass
class MetricElement(Dictable):
    _TYPE = ObjectSubType.METRIC

    id: str
    name: Optional[str] = None
    alias: Optional[str] = None

    def to_dict(self, camel_case: bool = True) -> dict:
        return {
            **super().to_dict(camel_case),
            'subType': self._TYPE.value,
        }


@dataclass
class MetricTemplateUnit(TemplateUnit):
    _TYPE = TemplateUnitType.METRICS
    _FROM_DICT_MAP = {
        'elements': [MetricElement.from_dict],
    }
    elements: list[MetricElement]


@dataclass
class Template(Dictable):
    _FROM_DICT_MAP = {
        'rows': [TemplateUnit.dispatch],
        'columns': [TemplateUnit.dispatch],
        'page_by': [TemplateUnit.dispatch],
    }
    rows: list[TemplateUnit] = field(default_factory=list)
    columns: list[TemplateUnit] = field(default_factory=list)
    page_by: list[TemplateUnit] = field(default_factory=list)



from typing import TYPE_CHECKING

from mstrio.utils.api_helpers import changeset_manager
from mstrio.utils.error_handlers import ErrorHandler
from mstrio.utils.helper import response_handler

if TYPE_CHECKING:
    from mstrio.connection import Connection


@ErrorHandler(
    err_msg='Selected project {name} does not exist or is not loaded.'
    ' Please load the project or select a valid project '
    'or create a new project using `create_new` method'
)
def get_project(
    connection,
    name,
    error_msg=None,
    throw_error=True,
    whitelist=[('ERR001', 500), ('ERR014', 403)],
):
    """Get a specific project that the authenticated user has access to.

    Args:
        connection: MicroStrategy REST API connection object
        name (string): MicroStrategy project name
        error_msg (string, optional): Custom Error Message for Error Handling
        whitelist(list): list of tuples of I-Server Error and HTTP errors codes
            respectively, which will not be handled
            i.e. whitelist = [('ERR001', 500),('ERR004', 404)]

    Returns:
        Complete HTTP response object.
    """
    return connection.get(url=f'{connection.base_url}/api/projects/{name}')


@ErrorHandler(err_msg='Error fetching list of available projects.')
def get_projects(connection, error_msg=None, whitelist=None):
    """Get a list of all projects that the authenticated user has access to.

    Args:
        connection: MicroStrategy REST API connection object
        error_msg (string, optional): Custom Error Message for Error Handling
        whitelist(list): list of tuples of I-Server Error and HTTP errors codes
            respectively, which will not be handled
            i.e. whitelist = [('ERR001', 500),('ERR004', 404)]

    Returns:
        Complete HTTP response object.
    """

    return connection.get(
        url=f'{connection.base_url}/api/projects', headers={'X-MSTR-ProjectID': None}
    )


def create_project(connection, body, error_msg=None):
    """Create a new project, either synchronously or asynchronously.

    Args:
        connection: MicroStrategy REST API connection object
        body: JSON-formatted definition of the dataset. Generated by
            `utils.formjson()`.
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    response = connection.post(
        url=connection.base_url + '/api/projects',
        headers={'X-MSTR-ProjectID': None},
        json=body,
    )
    if response.status_code != 201:
        if error_msg is None:
            error_msg = "Project could not be created."
        response_handler(response, error_msg)
    return response


@ErrorHandler(err_msg='Error getting project import quota for project with ID {id}')
def get_project_import_quota(connection, id, error_msg=None):
    """Get the amount of space, in MB, that can be used for the Data Import
    function for a specific project. This is the default value applied to all
    users.

    Args:
        connection: MicroStrategy REST API connection object
        id (string): Project id string
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    return connection.get(url=f'{connection.base_url}/api/projects/{id}/quotas')


@ErrorHandler(err_msg='Error setting import quota for project with ID {id}')
def set_project_import_quota(connection, id, body, error_msg=None):
    """Set the amount of space, in MB, that can be used for the Data Import
    function for a specific project.

    Args:
        connection: MicroStrategy REST API connection object
        id (string): Project id string
        body: JSON-formatted definition of the dataset. Generated by
            `utils.formjson()`.
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    return connection.put(
        url=f'{connection.base_url}/api/projects/{id}/quotas',
        headers={'X-MSTR-ProjectID': None},
        json=body,
    )


@ErrorHandler(
    err_msg='Error setting user {user_id} import quota for project with ID {id}'
)
def set_user_import_quota(connection, id, user_id, body, error_msg=None):
    """Set the amount of space, in MB, that can be used for the Data Import
    function for a specific user. The value provided is rounded to an integer.

    Args:
        connection: MicroStrategy REST API connection object
        id (string): Project id string
        user_id (string): User ID string
        body: JSON-formatted definition of the dataset. Generated by
            `utils.formjson()`.
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    return connection.put(
        url=f'{connection.base_url}/api/projects/{id}/users/{user_id}/quotas',
        headers={'X-MSTR-ProjectID': None},
        json=body,
    )


@ErrorHandler(err_msg='Error getting user import quota for project with ID {id}')
def get_user_import_quota(connection, id, user_id=None, error_msg=None):
    """Get a list of users for whom Data Import quotas have been set and their
    quotas.

    Args:
        connection: MicroStrategy REST API connection object
        id (string): Project id string
        user_id (string, optional): User id string
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    return connection.get(
        url=f'{connection.base_url}/api/projects/{id}/users/quotas',
        headers={'X-MSTR-ProjectID': None},
        params={'user_id': user_id},
    )


@ErrorHandler(err_msg='Error fetching project {id} settings configuration.')
def get_project_settings_config(connection, id, error_msg=None):
    """Get project settings configurations.

    Args:
        connection: MicroStrategy REST API connection object
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    return connection.get(
        url=f'{connection.base_url}/api/v2/projects/{id}/settings/config',
        headers={'X-MSTR-ProjectID': None},
    )


@ErrorHandler(err_msg='Error getting project settings for project with ID {id}')
def get_project_settings(connection, id, error_msg=None, whitelist=None):
    """Get project settings.

    Args:
        connection: MicroStrategy REST API connection object
        id (string): Project id string
        error_msg (string, optional): Custom Error Message for Error Handling
        whitelist(list): list of tuples of I-Server Error and HTTP errors codes
            respectively, which will not be handled
            i.e. whitelist = [('ERR001', 500),('ERR004', 404)]

    Returns:
        Complete HTTP response object.
    """
    return connection.get(
        url=f'{connection.base_url}/api/v2/projects/{id}/settings',
        headers={'X-MSTR-ProjectID': None},
    )


@ErrorHandler(err_msg='Error setting project settings for project with ID {id}')
def set_project_settings(connection, id, body, error_msg=None):
    """Set new project settings.

    Args:
        connection: MicroStrategy REST API connection object
        id (string, optional): Project id string
        body (dict): body of the request
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    return connection.put(
        url=f'{connection.base_url}/api/v2/projects/{id}/settings',
        headers={'X-MSTR-ProjectID': None},
        json=body,
    )


@ErrorHandler(err_msg='Error updating project settings for project with ID {id}')
def update_project_settings(connection, id, body, error_msg=None):
    """Update project settings.

    Args:
        connection: MicroStrategy REST API connection object
        id (string): Project id string
        body (dict): body of the request
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    return connection.patch(
        url=f'{connection.base_url}/api/v2/projects/{id}/settings',
        headers={'X-MSTR-ProjectID': None},
        json=body,
    )


@ErrorHandler(err_msg='Error getting engine setting for project with ID {id}')
def get_engine_settings(connection, id, error_msg=None):
    """Get available and current engine settings for a project.

    Args:
        connection: MicroStrategy REST API connection object
        id (string): Project id string
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    return connection.get(
        url=f'{connection.base_url}/api/projects/{id}/settings/engine',
        headers={'X-MSTR-ProjectID': None},
    )


@ErrorHandler(err_msg='Error getting project startup settings.')
def get_projects_on_startup(connection, error_msg=None, whitelist=None):
    """Get a list of projects along with the nodes on which they would be
    available when the iServer starts up.

    Args:
        connection: MicroStrategy REST API connection object
        error_msg (string, optional): Custom Error Message for Error Handling
        whitelist(list): list of tuples of I-Server Error and HTTP errors codes
            respectively, which will not be handled
            i.e. whitelist = [('ERR001', 500),('ERR004', 404)]

    Returns:
        Complete HTTP response object.
    """
    return connection.get(
        url=f'{connection.base_url}/api/projects/settings/onStartup',
        headers={'X-MSTR-ProjectID': None},
    )


@ErrorHandler(err_msg='Error updating project startup settings.')
def update_projects_on_startup(connection, body, error_msg=None, whitelist=None):
    """Update status of projects on iServer nodes at start up. You provide
    the request body as of list of replace operations to be performed on the
    value of array of nodes with the path URI containing the corresponding
    project id that needs to be updated.

    Args:
        connection: MicroStrategy REST API connection object
        body: {"operationList": [{
                    "op": "replace",
                    "path": "/projects/B7CA92F04B9FAE8D941C3E9B7E0CD754/nodes",
                    "value": ["env-183260laio2use1"]
                    }]
            }
        error_msg (string, optional): Custom Error Message for Error Handling
        whitelist(list): list of tuples of I-Server Error and HTTP errors codes
            respectively, which will not be handled
            i.e. whitelist = [('ERR001', 500),('ERR004', 404)]

    Returns:
        Complete HTTP response object.
    """
    return connection.patch(
        url=f'{connection.base_url}/api/projects/settings/onStartup',
        headers={'X-MSTR-ProjectID': None},
        json=body,
    )


@ErrorHandler(err_msg='Error getting VLDB settings for project with ID {id}')
def get_vldb_settings(connection: 'Connection', id: str, error_msg: str = None):
    """Get advanced VLDB settings for a project.

    Args:
        connection (Connection): MicroStrategy REST API connection object
        id (string): Project ID
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    return connection.get(
        url=f'{connection.base_url}/api/model/projects/{id}?showAdvancedProperties=true'
    )


@ErrorHandler(err_msg='Error updating VLDB settings for project with ID {id}')
def update_vldb_settings(
    connection: 'Connection', id: str, body: dict, error_msg: str = None
):
    """Update metadata of advanced VLDB settings for a project.

    Args:
        connection (Connection): MicroStrategy REST API connection object
        id (string): Project ID
        body (dict): JSON-formatted data used to update VLDB settings
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    with changeset_manager(connection) as changeset_id:
        return connection.put(
            url=f'{connection.base_url}/api/model/projects/{id}',
            headers={'X-MSTR-MS-Changeset': changeset_id},
            json=body,
        )


@ErrorHandler(
    err_msg='Error getting metadata of VLDB settings for project with ID {id}'
)
def get_applicable_vldb_settings(
    connection: 'Connection', id: str, error_msg: str = None
):
    """Get metadata of advanced VLDB settings for a project.

    Args:
        connection (Connection): MicroStrategy REST API connection object
        id (string): Project ID
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    return connection.get(
        url=f'{connection.base_url}/api/model/projects/{id}'
        '/applicableAdvancedProperties'
    )



from requests.exceptions import SSLError

from mstrio.utils.helper import exception_handler, response_handler


def server_status(connection):
    """
     Args:
        connection: MicroStrategy REST API connection object
    Returns:
        Complete HTTP response object
    """

    try:
        response = connection.get(
            skip_expiration_check=True, url=connection.base_url + '/api/status'
        )
    except SSLError:
        exception_handler(
            (
                "SSL certificate error.\nPlease double check that the link you "
                "are using comes from a trusted source. If you trust the URL "
                "provided please specify parameter 'ssl_verify' to 'False' in the "
                "'Connection' class.\n\nCheck readme for more details."
            ),
            SSLError,
        )

    if not response.ok:
        response_handler(response, "Failed to check server status")
    return response



"""This is the demo script to show how to manage intelligent (OLAP) cubes.

This script will not work without replacing parameters with real values.
Its basic goal is to present what can be done with this module and to
ease its usage.
"""

from mstrio.project_objects import list_olap_cubes, OlapCube
from mstrio.connection import get_connection

# Define a variable which can be later used in a script
PROJECT_NAME = $project_name  # Project to connect to

conn = get_connection(workstationData, project_name=PROJECT_NAME)

# list all OLAP Cubes
olap_cubes = list_olap_cubes(conn)

# Define a variable which can be later used in a script
CUBE_ID = $cube_id  # id for OlapCube object lookup

# get OLAP Cube by its id
olap_cube = OlapCube(conn, CUBE_ID)

# list available attributes, metrics and attribute forms which can be used for
# creation of a new OLAP Cube
OlapCube.available_metrics(conn)
OlapCube.available_attributes(conn)

# Define variables which can be later used in a script
ATTRIBUTE_ID = $attribute_id
ATTRIBUTE_NAME = $attribute_name
ATTRIBUTE = {'id': ATTRIBUTE_ID, 'name': ATTRIBUTE_NAME, 'type': 'attribute'}
METRIC_ID = $metric_id
METRIC_NAME = $metric_name
METRIC = {'id': METRIC_ID, 'name': METRIC_NAME, 'type': 'metric'}
CUBE_NAME = $cube_name
CUBE_DESCRIPTION = $cube_description
CUBE_FOLDER_ID = $cube_folder_id

# create new OLAP Cube
attributes = [ATTRIBUTE]
metrics = [METRIC]
new_olap_cube = OlapCube.create(
    conn,
    name=CUBE_NAME,
    description=CUBE_DESCRIPTION,
    folder_id=CUBE_FOLDER_ID,
    attributes=attributes,
    metrics=metrics
)

# Update attributes and metrics of a newly created OLAP Cube. When cube is not
# published yet, then it is possible to add/remove metrics or attributes.
# When cube is published it is possible to only change order of attributes or
# metrics which have already been added to cube.
attributes.append(ATTRIBUTE)
metrics.append(METRIC)
new_olap_cube.update(attributes=attributes, metrics=metrics)

# publish newly created and updated OLAP Cube
new_olap_cube.publish()

# refresh and show status of an OLAP Cube
new_olap_cube.refresh_status()
new_olap_cube.show_status()

# export sql view of an OLAP Cube
sql_view = new_olap_cube.export_sql_view()

# unpublish an OLAP Cube
new_olap_cube.unpublish()

# Delete an OLAP Cube. When `force` argument is set to `False` (default value)
# then deletion must be approved.
new_olap_cube.delete(force=True)



import json
from typing import Optional

from mstrio.connection import Connection
from mstrio.utils.api_helpers import (
    FuturesSessionWithRenewal,
    changeset_manager,
    unpack_information,
)
from mstrio.utils.datasources import (
    alter_conn_list_resp,
    alter_conn_resp,
    alter_instance_list_resp,
    alter_instance_resp,
    alter_patch_req_body,
)
from mstrio.utils.error_handlers import ErrorHandler
from mstrio.utils.helper import IServerError, exception_handler, response_handler


@ErrorHandler(err_msg='Error getting available DBMSs.')
def get_available_dbms(connection, error_msg=None):
    """Get information for all available database management systems (DBMSs).

    Args:
        connection: MicroStrategy REST API connection object
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. HTTP STATUS 200/400
    """
    url = f"{connection.base_url}/api/dbobjects/dbmss"
    return connection.get(url=url)


@ErrorHandler(err_msg='Error getting available database drivers.')
def get_available_db_drivers(connection, error_msg=None):
    """Get information for all available database drivers.

    Args:
        connection: MicroStrategy REST API connection object
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. HTTP STATUS 200/400
    """
    url = f"{connection.base_url}/api/dbobjects/drivers"
    return connection.get(url=url)


def get_datasource_instance(connection, id, error_msg=None):
    """Get information for a specific database source.

    Args:
        connection: MicroStrategy REST API connection object
        id (string): ID
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. HTTP STATUS 200/400
    """
    url = f"{connection.base_url}/api/datasources/{id}"
    response = connection.get(url=url)
    if not response.ok:
        if error_msg is None:
            error_msg = f"Error getting Datasource Instance with ID: {id}"
        response_handler(response, error_msg)
    response = alter_instance_resp(response)
    return response


@ErrorHandler(err_msg='Error deleting Datasource Instance with ID {id}')
def delete_datasource_instance(connection, id, error_msg=None):
    """Delete a specific database source based on id.

    Args:
        connection: MicroStrategy REST API connection object
        id (string): ID
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. HTTP STATUS 204/400
    """
    url = f"{connection.base_url}/api/datasources/{id}"
    return connection.delete(url=url)


def update_datasource_instance(connection, id, body, error_msg=None):
    """Update a specific database source based on id.

    Args:
        connection: MicroStrategy REST API connection object
        id (string): ID
        body: update operation info
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. HTTP STATUS 200/400
    """
    url = f"{connection.base_url}/api/datasources/{id}"
    for op_dict in body["operationList"]:
        op_dict = alter_patch_req_body(
            op_dict, "/datasourceConnection", "/databaseConnectionId"
        )
        op_dict = alter_patch_req_body(
            op_dict, "/primaryDatasource", "/databasePrimaryDatasourceId"
        )
        op_dict = alter_patch_req_body(
            op_dict, "/dataMartDatasource", "/databaseDataMartDatasourceId"
        )
        alter_patch_req_body(op_dict, "/dbms", "/dbmsId")
    response = connection.patch(url=url, json=body)
    if not response.ok:
        if error_msg is None:
            error_msg = f"Error updating Datasource Instance with ID: {id}"
        response_handler(response, error_msg)
    return response


def create_datasource_instance(connection, body, error_msg=None):
    """Create a specific database source.

    Args:
        connection: MicroStrategy REST API connection object
        id (string): ID
        body: Datasource info
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. HTTP STATUS 201/400
    """
    url = f"{connection.base_url}/api/datasources"
    response = connection.post(url=url, json=body)
    if not response.ok:
        if error_msg is None:
            name = body.get("name", "NA")
            error_msg = f"Error creating Datasource Instance: {name}"
        response_handler(response, error_msg)
    response = alter_instance_resp(response)
    return response


@ErrorHandler(err_msg='Error getting the namespaces for datasource with ID: {id}.')
def get_datasource_namespaces(
    connection: "Connection",
    id: str,
    project_id: Optional[str] = None,
    refresh: Optional[bool] = None,
    fields: Optional[str] = None,
    error_msg: Optional[str] = None,
):
    """Get namespaces for a specific datasource.

    Args:
        connection: MicroStrategy REST API connection object
        id (str): Database ID
        project_id (str, optional): Project ID
        refresh (bool, optional): Force refresh
        fields (str, optional): A whitelist of top-level fields separated by
            commas. Allow the client to selectively retrieve fields in the
            response.
        error_msg (str, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. Expected status is 200.
    """
    if project_id is None:
        connection._validate_project_selected()
        project_id = connection.project_id

    return connection.get(
        url=f"{connection.base_url}/api/datasources/{id}/catalog/namespaces",
        headers={
            'X-MSTR-ProjectID': project_id,
        },
        params={
            'refresh': refresh,
            'fields': fields,
        },
    )


def get_datasource_namespaces_async(
    session: FuturesSessionWithRenewal,
    connection: "Connection",
    id: str,
    project_id: Optional[str] = None,
    refresh: Optional[bool] = None,
    fields: Optional[str] = None,
):
    return session.get(
        url=f"{connection.base_url}/api/datasources/{id}/catalog/namespaces",
        headers={
            'X-MSTR-ProjectID': project_id,
        },
        params={
            'refresh': refresh,
            'fields': fields,
        },
    )


def get_datasource_instances(
    connection, ids=None, database_type=None, project=None, error_msg=None
):
    """Get information for all database sources.

    Args:
        connection: MicroStrategy REST API connection object
        ids: list of datasources ids
        database_type: list of types (names) of databases
        project: id (str) of a project or instance of an Project class
            to search for the datasource instances in. When provided, both
            `ids` and `database_types` are ignored. By default `None`.
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. HTTP STATUS 200/400
    """
    if project:
        url = f"{connection.base_url}/api/projects/{project}/datasources"
        response = connection.get(url=url)
    else:
        database_type = None if database_type is None else database_type.join(",")
        ids = None if ids is None else ids.join(",")
        url = f"{connection.base_url}/api/datasources"
        response = connection.get(
            url=url, params={'id': ids, 'database.type': database_type}
        )
    if not response.ok:
        res = response.json()
        if project and res.get("message") == "HTTP 404 Not Found":
            # aka project based endpoint not supported
            # try without filtering
            warning_msg = (
                "get_datasource_instances() warning: filtering by Project "
                "is not yet supported on this version of the I-Server. "
                "Returning all values."
            )
            exception_handler(warning_msg, Warning)
            return get_datasource_instances(
                connection=connection,
                ids=ids,
                database_type=database_type,
                error_msg=error_msg,
            )
        if not error_msg:
            if (
                project
                and res.get('code') == "ERR006"
                and "not a valid value for Project ID" in res.get('message')
            ):
                error_msg = f"{project} is not a valid Project class instance or ID"
                raise ValueError(error_msg)
            error_msg = "Error getting Datasource Instances"
            if project:
                error_msg += f" within `{project}` Project"
        response_handler(response, error_msg)
    response = alter_instance_list_resp(response)
    return response


def get_datasource_connections(connection, error_msg=None):
    """Get information for all datasource connections.

    Args:
        connection: MicroStrategy REST API connection object
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. HTTP STATUS 200/400
    """
    url = f"{connection.base_url}/api/datasources/connections"

    response = connection.get(url=url)
    if not response.ok:
        if error_msg is None:
            error_msg = "Error getting Datasource Connections"
        response_handler(response, error_msg)
    response = alter_conn_list_resp(response)
    return response


def get_datasource_connection(connection, id, error_msg=None):
    """Get a datasource connection for given id.

    Args:
        connection: MicroStrategy REST API connection object
        id (string): ID
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. HTTP STATUS 200/400
    """
    url = f"{connection.base_url}/api/datasources/connections/{id}"
    response = connection.get(url=url)
    if not response.ok:
        if error_msg is None:
            error_msg = (
                f"Error getting Datasource Connection with ID: {id}. "
                f"Check if it is not embedded Datasource Connection."
            )
        response_handler(response, error_msg)
    response = alter_conn_resp(response)
    return response


def update_datasource_connection(connection, id, body, error_msg=None):
    """Update a datasource connection based on id.

    Args:
        connection: MicroStrategy REST API connection object
        id (string): ID
        body: update operation info
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. HTTP STATUS 200/400
    """
    url = f"{connection.base_url}/api/datasources/connections/{id}"
    for op_dict in body["operationList"]:
        alter_patch_req_body(op_dict, "/datasourceLogin", "/databaseLoginId")
    response = connection.patch(url=url, json=body)
    if not response.ok:
        if error_msg is None:
            error_msg = f"Error updating Datasource Connection with ID: {id}"
        response_handler(response, error_msg)
    return response


@ErrorHandler(err_msg='Error deleting Datasource Connection with ID {id}')
def delete_datasource_connection(connection, id):
    """Delete a datasource connection based on id.

    Args:
        connection: MicroStrategy REST API connection object
        id (string): ID
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. HTTP STATUS 204/400
    """
    url = f"{connection.base_url}/api/datasources/connections/{id}"
    return connection.delete(url=url)


def create_datasource_connection(connection, body, error_msg=None):
    """Create a specific database connection.

    Args:
        connection: MicroStrategy REST API connection object
        body: Datasource Connection info
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. HTTP STATUS 201/400
    """
    url = f"{connection.base_url}/api/datasources/connections"
    response = connection.post(url=url, json=body)
    if not response.ok:
        if error_msg is None:
            name = body.get("name", "NA")
            error_msg = f"Error creating Datasource connection: {name}"
        response_handler(response, error_msg)
    response = alter_conn_resp(response)
    return response


@ErrorHandler(err_msg='Error testing Datasource connection.')
def test_datasource_connection(connection, body, error_msg=None):
    """Test a datasource connection. Either provide a connection id, or the
    connection parameters within connection object.

    Args:
        connection: MicroStrategy REST API connection object.
        body: Datasource Connection info.
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. HTTP STATUS 204/400
    """
    url = f"{connection.base_url}/api/datasources/connections/test"
    return connection.post(url=url, json=body)


@ErrorHandler(err_msg='Error fetching connection mappings.')
def get_datasource_mappings(
    connection: Connection,
    default_connection_map: Optional[bool] = False,
    project_id: Optional[str] = None,
    error_msg: Optional[str] = None,
):
    """Get information for all connection mappings.

    Args:
        connection: MicroStrategy REST API connection object
        default_connection_map (bool, optional): If True will get the default
            connection mappings for a project. Requires `project_id`
            parameter. Default False.
        project_id: The project_id, required only for default connection
            mappings.
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. Expected status is 200.
    """
    response = connection.get(
        url=f"{connection.base_url}/api/datasources/mappings",
        params={
            "defaultConnectionMap": default_connection_map,
            "projectId": project_id,
        },
    )

    if default_connection_map and not response.ok and response.status_code == 404:
        response.status_code = 200
        response._content = json.dumps({'mappings': []}).encode('utf-8')

    return response


@ErrorHandler(err_msg='Error fetching connection mapping with ID {id}.')
def get_datasource_mapping(
    connection: Connection,
    id=str,
    default_connection_map: Optional[bool] = False,
    project_id: Optional[str] = None,
    error_msg: Optional[str] = None,
):
    """Get information about specific connection mapping.

    Args:
        connection: MicroStrategy REST API connection object
        id: ID of the mapping
        default_connection_map (bool, optional): If True will get the default
            connection mappings for a project. Requires `project_id`
            parameter. Default False.
        project_id: The project_id, required only for default connection
            mappings.
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. Expected status is 200.
    """

    response = get_datasource_mappings(
        connection=connection,
        default_connection_map=default_connection_map,
        project_id=project_id,
        error_msg=error_msg,
    )

    # Faking get single resource endpoint. Only 'list all' available on REST
    if response.ok:
        response_json = response.json()

        try:
            mappings = [
                mapping for mapping in response_json['mappings'] if mapping["id"] == id
            ]

            mapping_data = mappings[0]
            mapping_data['ds_connection'] = mapping_data.pop('connection')
        except LookupError:
            raise IServerError(message="Connection Mapping not found", http_code=None)

        response.encoding, response._content = 'utf-8', json.dumps(mapping_data).encode(
            'utf-8'
        )
    return response


@ErrorHandler(err_msg='Error creating connection mapping.')
def create_datasource_mapping(
    connection: Connection, body, error_msg: Optional[str] = None
):
    """Create a new connection mapping.

    Args:
        connection: MicroStrategy REST API connection object
        body: Datasource Connection Map creation info.
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. Expected status is 201.
    """
    url = f"{connection.base_url}/api/datasources/mappings"
    return connection.post(url=url, json=body)


@ErrorHandler(err_msg='Error deleting connection mapping with ID {id}')
def delete_datasource_mapping(
    connection: Connection, id: str, error_msg: Optional[str] = None
):
    """Delete a connection mapping based on id.

    Args:
        connection: MicroStrategy REST API connection object
        id (string): ID of the mapping meant to be deleted.
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. Expected status is 204.
    """
    url = f"{connection.base_url}/api/datasources/mappings/{id}"
    return connection.delete(url=url)


@ErrorHandler(err_msg="Error getting Datasource logins.")
def get_datasource_logins(connection: Connection, error_msg: Optional[str] = None):
    """Get information for all datasource logins.

    Args:
        connection: MicroStrategy REST API connection object
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. Expected status is 200.
    """
    url = f"{connection.base_url}/api/datasources/logins"
    return connection.get(url=url)


@ErrorHandler(err_msg='Error creating Datasource login.')
def create_datasource_login(
    connection: Connection, body, error_msg: Optional[str] = None
):
    """Create a new datasource login.

    Args:
        connection: MicroStrategy REST API connection object
        body: Datasource login creation info.
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. Expected status is 201.
    """
    url = f"{connection.base_url}/api/datasources/logins"
    return connection.post(url=url, json=body)


@ErrorHandler(err_msg='Error getting Datasource login with ID {id}')
def get_datasource_login(
    connection: Connection, id: str, error_msg: Optional[str] = None
):
    """Get datasource login for a specific id.

    Args:
        connection: MicroStrategy REST API connection object
        id: ID of the login
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. Expected status is 200.
    """
    url = f"{connection.base_url}/api/datasources/logins/{id}"
    return connection.get(url=url)


@ErrorHandler(err_msg='Error deleting Datasource login with ID {id}')
def delete_datasource_login(
    connection: Connection, id: str, error_msg: Optional[str] = None
):
    """Delete a datasource login.

    Args:
        connection: MicroStrategy REST API connection object
        id: ID of the login
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. Expected status is 204.
    """
    url = f"{connection.base_url}/api/datasources/logins/{id}"
    return connection.delete(url=url)


@ErrorHandler(err_msg='Error updating Datasource login with ID {id}')
def update_datasource_login(
    connection: Connection, id: str, body, error_msg: Optional[str] = None
):
    """Update a datasource login.

    Args:
        connection: MicroStrategy REST API connection object
        id: ID of the login
        body: Datasource Connection Map creation info.
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. Expected status is 200.
    """
    url = f"{connection.base_url}/api/datasources/logins/{id}"
    return connection.patch(url=url, json=body)


@ErrorHandler(err_msg="Error getting table columns for table: {table_id}")
def get_table_columns(
    connection: Connection,
    datasource_id: str,
    namespace_id: str,
    table_id: str,
    error_msg: Optional[str] = None,
):
    url = (
        f"{connection.base_url}/api/datasources/{datasource_id}/catalog/namespaces/"
        f"{namespace_id}/tables/{table_id}"
    )
    return connection.get(url, headers={"X-MSTR-ProjectID": connection.project_id})


@unpack_information
@ErrorHandler(
    err_msg='Error converting Datasource embedded connection from DSN to DSN-less'
)
def convert_ds_dsn(
    connection: Connection, datasource_id: str, error_msg: Optional[str] = None
):
    """Convert datasource embedded connection from DSN to DSN-less format
    connection string and update the object to metadata.

    Args:
        connection: MicroStrategy REST API connection object
        datasource_id (string) : Datasource id
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        HTTP response object with updated embedded connection data.
        Expected status is 200.
    """
    url = f"{connection.base_url}/api/datasources/{datasource_id}/conversion"
    return connection.post(url=url)


@unpack_information
@ErrorHandler(
    err_msg='Error converting Datasource connection object from DSN to DSN-less'
)
def convert_connection_dsn(
    connection: Connection, ds_connection_id: str, error_msg: Optional[str] = None
):
    """Convert datasource connection from DSN to DSN-less format connection
    string and update the object to metadata.

    Args:
        connection: MicroStrategy REST API connection object
        ds_connection_id (string) : Datasource connection object id
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        HTTP response object with updated object data. Expected status is 200.
    """
    url = (
        f"{connection.base_url}/api/datasources/connections/{ds_connection_id}"
        f"/conversion"
    )
    return connection.post(url=url)


@ErrorHandler(err_msg='Error getting VLDB settings for datasource with ID: {id}')
def get_vldb_settings(connection: 'Connection', id: str, error_msg: str = None):
    """Get advanced VLDB settings for a datasource.

    Args:
        connection (Connection): MicroStrategy REST API connection object
        id (string): Datasource ID
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """

    return connection.get(
        url=f'{connection.base_url}/api/model/datasources/{id}'
        '?showAdvancedProperties=true'
    )


@ErrorHandler(err_msg='Error updating VLDB settings for datasource with ID {id}')
def update_vldb_settings(
    connection: 'Connection', id: str, body: dict, error_msg: str = None
):
    """Update metadata of advanced VLDB settings for a datasource.

    Args:
        connection (Connection): MicroStrategy REST API connection object
        id (string): Datasource ID
        body (dict): JSON-formatted data used to update VLDB settings
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """

    with changeset_manager(connection) as changeset_id:
        return connection.put(
            url=f'{connection.base_url}/api/model/datasources/{id}',
            json=body,
            headers={'X-MSTR-MS-Changeset': changeset_id},
        )


@ErrorHandler(
    err_msg='Error getting metadata of VLDB settings for datasource with ID {id}'
)
def get_applicable_vldb_settings(
    connection: 'Connection', id: str, error_msg: str = None
):
    """Get metadata of advanced VLDB settings for a datasource.

    Args:
        connection (Connection): MicroStrategy REST API connection object
        id (string): Datasource ID
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """

    return connection.get(
        url=f'{connection.base_url}/api/model/datasources/{id}'
        '/applicableAdvancedProperties'
    )



"""List user privileges for all active users."""

from typing import List

from mstrio.connection import Connection, get_connection
from mstrio.users_and_groups import list_users


def list_active_user_privileges(connection: "Connection") -> List[dict]:
    """List user privileges for all active users.

    Args:
        connection: MicroStrategy connection object returned by
            `connection.Connection()`

    Returns:
        list of dicts where each of them is in given form:
        {
            'id' - id of user
            'name' - name of user
            'username' - username of user
            'privileges' - list of privileges of user
        }
    """
    all_users = list_users(connection=connection)
    active_users = [u for u in all_users if u.enabled]
    privileges_list = []
    for usr in active_users:
        p = {
            'id': usr.id,
            'name': usr.name,
            'username': usr.username,
            'privileges': usr.privileges,
        }
        print(f"{p['name']} ({p['username']}) ", flush=True)
        for prvlg in p['privileges']:
            print("\t" + prvlg['privilege']['name'], flush=True)
        privileges_list.append(p)
    return privileges_list


# connect to environment without providing user credentials
# variable `workstationData` is stored within Workstation
conn = get_connection(workstationData, 'MicroStrategy Tutorial')

# list privileges for all enabled users
list_active_user_privileges(conn)



from enum import auto

from mstrio.utils.enum_helper import AutoName


class RefreshPolicy(AutoName):
    ADD = auto()
    DELETE = auto()
    UPDATE = auto()
    UPSERT = auto()
    REPLACE = auto()



from mstrio.api import objects as objects_api


def get_info(
    connection,
    id: str,
    object_type: int,
    project_id: str = None,
) -> dict:
    """Get information for a specific object in a specific project; if you do
    not specify a project ID, you get information for the object just in the
    non-project area.

    You identify the object with the object ID and object type. You specify
    the object type as a query parameter; possible values for object type are
    provided in EnumDSSXMLObjectTypes.

    Args:
        connection(object): MicroStrategy connection object returned by
            `connection.Connection()`.
        id (str): Object ID
        object_type (int): One of EnumDSSXMLObjectTypes.
        project_id (str): ID of a project in which the object is located.

    Returns:
        dict
    """
    return objects_api.get_object_info(connection, id, object_type, project_id).json()



"""This is the demo script to show how to manage documents. Its basic goal is
to present what can be done with this module and to ease its usage.
"""

from mstrio.connection import get_connection
from mstrio.project_objects.document import (
    Document, list_documents, list_documents_across_projects
)

# Define a variable which can be later used in a script
PROJECT_NAME = $project_name  # Insert name of project here

conn = get_connection(workstationData, PROJECT_NAME)

# Define variables which can be later used in a script
PROJECT_ID = $project_id # Insert project ID here
DOCUMENT_NAME = $document_name # Insert name of document on which you want to perform actions

# Document management
# Listing documents with different conditions
list_of_all_documents = list_documents(connection=conn)
print(list_of_all_documents)
list_of_all_documents_as_dicts = list_documents(connection=conn, to_dictionary=True)
print(list_of_all_documents_as_dicts)
list_of_all_documents_as_dataframes = list_documents(connection=conn, to_dataframe=True)
print(list_of_all_documents_as_dataframes)
list_of_documents_on_project_by_id = list_documents(connection=conn, project_id=PROJECT_ID)
print(list_of_documents_on_project_by_id)
list_of_documents_on_project_by_name = list_documents(connection=conn, project_name=PROJECT_NAME)
print(list_of_documents_on_project_by_name)
list_of_documents_by_name = list_documents(connection=conn, name=DOCUMENT_NAME)
print(list_of_documents_by_name)

# List of documents across projects, with examples of different conditions
list_of_documents_across_projects = list_documents_across_projects(connection=conn)
print(list_of_documents_across_projects)
list_of_documents_as_dataframes = list_documents_across_projects(connection=conn, to_dataframe=True)
print(list_of_documents_as_dataframes)
list_of_documents_as_dicts = list_documents_across_projects(connection=conn, to_dictionary=True)
print(list_of_documents_as_dicts)

# Define a variable which can be later used in a script
DOCUMENT_ID = $document_id  # Insert ID of document on which you want to perform actions

# Get single document by id
document = Document(connection=conn, id=DOCUMENT_ID)

# Get single document by name
document = Document(connection=conn, id=DOCUMENT_NAME)

# Listing properties of document
properties = document.list_properties()
print(properties)

# Define variables which can be later used in a script
DOCUMENT_NEW_NAME = $document_new_name  # Insert new name of edited document here
DOCUMENT_NEW_DESCRIPTION = $document_new_description  # Insert new description of edited document
FOLDER_ID = $folder_id  # Insert folder ID here here

# Alter document
document.alter(name=DOCUMENT_NEW_NAME)
document.alter(description=DOCUMENT_NEW_DESCRIPTION)
document.alter(folder_id=FOLDER_ID)

# Define a variable which can be later used in a script
USER_ID = $user_id  # Insert user ID here

# Publish document
document.publish()
document.publish(recipients=USER_ID)

# Unpublish document
document.unpublish()

# Listing schedules of document
schedules = document.list_available_schedules()
print(schedules)
schedules_as_dicts = document.list_available_schedules(to_dictionary=True)
print(schedules_as_dicts)

# Share document with given user
document.share_to(users=USER_ID)

# Listing cubes used by this document
cubes = document.get_connected_cubes()
print(cubes)

# Get list of document cache, with examples of different conditions
list_of_all_document_cache = Document.list_caches(connection=conn)
print(list_of_all_document_cache)
list_of_limited_document_cache = Document.list_caches(connection=conn, limit=5)
print(list_of_limited_document_cache)

# Define a variable which can be later used in a script
CACHE_ID = $cache_id  # Insert ID of cache on which you want to perform actions

list_of_document_cache_by_id = Document.list_caches(connection=conn, id=CACHE_ID)
print(list_of_document_cache_by_id)
list_of_document_caches_as_dicts = Document.list_caches(connection=conn, to_dictionary=True)
print(list_of_document_caches_as_dicts)

# Helper function to list all cache status
def show_caches_status():
    caches = Document.list_caches(conn)
    for cache in caches:
        print(f'Cache ID: {cache.id}, {cache.status}')

# List cache properties
cache = Document.list_caches(connection=conn)[0]
properties = cache.list_properties()
print(properties)

# Define a variable which can be later used in a script
OTHER_CACHE_ID = $other_cache_id  # Insert ID of cache on which you want to perform actions

# Unload multiple document caches
Document.unload_caches(connection=conn, cache_ids=[CACHE_ID, OTHER_CACHE_ID])
show_caches_status()

# Define variables which can be later used in a script
USER_NAME = $user_name # Insert user name here
CACHE_STATUS = $cache_status # Insert cache status here

# Unload all document caches with filter examples
Document.unload_all_caches(connection=conn)
show_caches_status()
Document.unload_all_caches(connection=conn, owner=USER_NAME)
show_caches_status()
Document.unload_all_caches(connection=conn, status=CACHE_STATUS)
show_caches_status()

# Load multiple document caches
Document.load_caches(connection=conn, cache_ids=[CACHE_ID, OTHER_CACHE_ID])
show_caches_status()

# Load all document caches with filter examples
Document.load_all_caches(connection=conn)
show_caches_status()
Document.load_all_caches(connection=conn, owner=USER_NAME)
show_caches_status()
Document.load_all_caches(connection=conn, status=CACHE_STATUS)
show_caches_status()

# Delete multiple document caches
Document.delete_caches(connection=conn, cache_ids=[CACHE_ID, OTHER_CACHE_ID], force=True)

# Delete all document caches with filter examples
Document.delete_all_caches(connection=conn, force=True)
Document.delete_all_caches(connection=conn, owner=USER_NAME, force=True)
Document.delete_all_caches(connection=conn, status=CACHE_STATUS, force=True)
